from linora.image._image_position import *

class ImagePositionAug(object):
    def __init__(self, image=None):
        self.image = image
    
    def flip_up_left(self, p=None):
        """Randomly flip an image (up to left).
    
        Args:
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = flip_up_left(self.image, p)
        return self
    
    def flip_up_right(self, p=None):
        """Randomly flip an image (up to right).
    
        Args:
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = flip_up_right(self.image, p)
        return self
    
    def flip_left_right(self, p=None):
        """Randomly flip an image (left to right).
    
        Args:
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = flip_left_right(self.image, p)
        return self
    
    def flip_up_down(self, p=None):
        """Randomly flip an image (up to down).
    
        Args:
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = flip_up_down(self.image, p)
        return self
    
    def rotate(self, angle, expand=False, center=(0.5,0.5), translate=(0,0), fillcolor=None, p=None):
        """Returns a rotated copy of this image. 

        This method returns a copy of this image, rotated the given number of degrees counter clockwise around its centre.

        Args:
            angle: In degrees counter clockwise, angle in [-180, 180].
                   if int or float, rotation angle.
                   if list or tuple, randomly picked in the interval `[angle[0], angle[1])` value.
            expand: if true, expands the output image to make it large enough to hold the entire rotated image. 
                    if false, make the output image the same size as the input image. 
                    Note that the expand flag assumes rotation around the center and no translation.
                    if value is None, then the function is random.
            center: center of rotation, xaxis and yaxis in [0,1], default is the center of the image.
                    if int or float, xaxis=yaxis,
                    if 2-tuple, (xaxis, yaxis), 
                    if 4-tuple, xaxis in (center[0], center[1]) and yaxis in (center[2], center[3]).
            translate: post-rotate translation, xoffset and yoffset in [-1,1], see la.image.translate method.
                       if int or float, xoffset=yoffset,
                       if 2-tuple, (xoffset, yoffset), 
                       if 4-tuple, xoffset in (translate[0], translate[1]) and  yoffset in (translate[2], translate[3]).
            fill_color: color for area outside, int or str or tuple or la.image.RGBMode, rgb color.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = rotate(self.image, angle, expand, center, translate, fillcolor, p)
        return self
    
    def translate(self, xoffset=(-0.5,0.5), yoffset=None, fill_color=None, p=None):
        """Returns a translate copy of this image. 

        Args:
            xoffset: [-1, 1], int or float, width offset.
                     if list or tuple, randomly picked in the interval `[xoffset[0], xoffset[1])`.
            yoffset: [-1, 1], int or float, height offset.
                     if list or tuple, randomly picked in the interval `[yoffset[0], yoffset[1])`.
            fill_color: int or str or tuple or la.image.RGBMode, rgb color.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = translate(self.image, xoffset, yoffset, fillcolor, p)
        return self
    
    def offset(self, xoffset, yoffset=None, p=None):
        """Returns a copy of the image where data has been offset by the given distances.

        Data wraps around the edges. If ``yoffset`` is omitted, it is assumed to be equal to ``xoffset``.

        Args:
            xoffset: int or list ot tuple, The horizontal distance.
                     if tuple or list, randomly picked in the interval `[xoffset[0], xoffset[1])`
            yoffset: int or list ot tuple, The vertical distance. 
                     If omitted, both distances are set to the same value.
                     if tuple or list, randomly picked in the interval `[yoffset[0], yoffset[1])`
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = offset(self.image, xoffset, yoffset, p)
        return self
    
    def pad(self, pad_value, pad_color=None, p=None):
        """Add border to the image

        Args:
            pad_value: int or list or tuple, if int, pad same value with border, 
                       if list or tuple, len(pad_value)==2, left, top = right, bottom = pad_value
                       if list or tuple, len(pad_value)==4, left, top, right, bottom = pad_value
            pad_color: str or tuple or list or la.image.RGBMode, fill RGB color value, 
                       if str, hexadecimal color;
                       if len(pad_color) == 2, left_color, top_color = right_color, bottom_color = pad_color
                       if len(pad_color) == 3, left_color = top_color = right_color = bottom_color = pad_color
                       if len(pad_color) == 4, left_color, top_color, right_color, bottom_color = pad_color
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = pad(self.image, pad_value, pad_color, p)
        return self
    
    def channel_shuffle(self, p=None):
        """Random shuffle image channel.

        Args:
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = channel_shuffle(self.image, p)
        return self
    
    def perspective(self, distortion_scale, fill_color=None, p=None):
        """Performs a random perspective transformation of the given image with a given probability. 

        Args:
            distortion_scale: float, argument to control the degree of distortion and ranges from 0 to 1.
            fill_color: int or str or tuple or la.image.RGBMode, rgb color.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = perspective(self.image, distortion_scale, fill_color, p)
        return self
    
    def affine(self, angle=(-180, 180), center=(0.5,0.5), translate=(0, 0), scale=1., shear=(0,0), fill_color=None, p=None):
        """Apply affine transformation on the image keeping image center invariant.

        Args:
            angle: int or float, rotation angle in degrees between -180 and 180. Set to 0 to deactivate rotations.
                   if list or tuple, randomly picked in the interval `[angle[0], angle[1])`.
            center: center of rotation, xaxis and yaxis in [0,1], default is the center of the image.
                    if int or float, xaxis=yaxis,
                    if 2-tuple, (xaxis, yaxis), 
                    if 4-tuple, xaxis in (center[0], center[1]) and yaxis in (center[2], center[3]).
            translate: post-rotate translation, xoffset and yoffset in [-1,1], see la.image.translate method.
                       if int or float, xoffset=yoffset,
                       if 2-tuple, (xoffset, yoffset), 
                       if 4-tuple, xoffset in (translate[0], translate[1]) and  yoffset in (translate[2], translate[3]).
            scale: float, scaling factor interval, should be positive.
                   if list or tuple, randomly picked in the interval `[scale[0], scale[1])`.
            shear: Range of degrees to select from, xoffset and yoffset in [-360,360]. 
                   if int or float, xoffset=yoffset,
                   if 2-tuple, (xoffset, yoffset), 
                   if 4-tuple, xoffset in (shear[0], shear[1]) and  yoffset in (shear[2], shear[3]).
            fill_color: int or str or tuple or la.image.RGBMode, rgb color. 
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = affine(self.image, angle, center, translate, scale, shear, fill_color, p)
        return self
    
    def shear(self, xoffset=(-90, 90), yoffset=None, fill_color=None, p=None):
        """Apply affine shear on the image.

        Args:
            xoffset: int or list ot tuple, The horizontal degrees, xoffset in [-360,360].
                     if tuple or list, randomly picked in the interval `[xoffset[0], xoffset[1])`
            yoffset: int or list ot tuple, The vertical degrees, yoffset in [-360,360]. 
                     If omitted, both distances are set to the same value.
                     if tuple or list, randomly picked in the interval `[yoffset[0], yoffset[1])`
            fill_color: int or str or tuple or la.image.RGBMode, rgb color. 
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = shear(self.image, xoffset, yoffset, fill_color, p)
        return self
    
    def rescale(self, xscale=(0.5,1.5), yscale=(0.5,1.5), fill_color=None, p=None):
        """Apply scaling on the y-axis to input data.

        Args:
            xscale: if int or float, width expansion and contraction, xscale should be positive.
                    if tuple or list, randomly picked in the interval `[xscale[0], xscale[1])`.
            yscale: if int or float, height expansion and contraction, yscale should be positive.
                    if tuple or list, randomly picked in the interval `[yscale[0], yscale[1])`.
            fill_color: int or str or tuple or la.image.RGBMode, rgb color.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = rescale(self.image, xscale, yscale, fill_color, p=p)
        return self
    
    def jigsaw(self, size=(10,10), prob=0.1, p=None):
        """Move cells within images similar to jigsaw patterns.

        Args:
            size: if int or float, xsize=ysize, numbers of jigsaw.
                  if 2-tuple, (xsize, ysize), 
                  if 4-tuple, xsize in (size[0], size[1]) and  ysize in (size[2], size[3]).
            prob: probability of every jigsaw being changed.
            p: probability that the image does this. Default value is 1.
        """
        if self._max_aug_nums>0:
            if self._nums>self._max_aug_nums:
                return self
            self._nums += 1
        if p is None:
            p = self._p
        self.image = jigsaw(self.image, size, prob, p)
        return self


import numpy as np


__all__ = ['numerical_binarizer', 'numerical_bucketized', 'numerical_cyclical',
           'numerical_combine', 'numerical_math', 'numerical_padding', 'numerical_outlier',
           'numerical_relative']


def numerical_binarizer(feature, mode=0, method='mean', name=None, config=None):
    """Encode labels with value between 0 and 1.
    
    Args:
        feature: pd.Series, sample feature.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        method: default 'mean', one of 'mean' or 'median' or float or int.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'method':method, 'name':feature.name if name is None else name},
                  'type':'numerical_binarizer', 'variable':feature.name, 'keep':keep}
        if method=='mean':
            config['param']['feature_scale'] = feature.mean()
        elif method=='median':
            config['param']['feature_scale'] = feature.median()
        else:
            config['param']['feature_scale'] = method
    if mode==2:
        return config
    else:
        scale = config['param']['feature_scale']
        t = (feature.clip(upper=scale).replace({scale:scale+0.1}).clip(scale)
             .replace({scale:0, scale+0.1:1}).astype('int8').rename(config['param']['name']))
        return t if mode else (t, config)


def numerical_bucketized(feature, boundaries, mode=0, miss_pad=-1, score=None, miss_score=None, method=1, name=None, config=None):
    """feature bucket.
    
    if method is True:
        Buckets include the right boundary, and exclude the left boundary. 
        Namely, boundaries=[0., 1., 2.] generates buckets (-inf, 0.], (0., 1.], (1., 2.], and (2., +inf).
    else:
        Buckets include the left boundary, and exclude the right boundary. 
        Namely, boundaries=[0., 1., 2.] generates buckets (-inf, 0.), [0., 1.), [1., 2.), and [2., +inf).
        
    Args:
        feature: pd.Series, sample feature.
        boundaries: list, A sorted list or tuple of floats specifying the boundaries.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        miss_pad: int, default -1, feature fillna value.
        score: None, A score list or tuple of floats specifying the boundaries.
        miss_score: int or float, None, score fillna value.
        method: True.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid, but `boundaries` must be passed in.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'boundaries':boundaries, 'miss_pad':miss_pad,
                           'score':score, 'miss_score':miss_score, 'method':method,
                           'name':feature.name if name is None else name},
                  'type':'numerical_bucketized', 'variable':feature.name}
    if mode==2:
        return config
    else:
        t = feature.copy()
        bound = sorted(config['param']['boundaries'])
        if config['param']['method']:
            for i in range(len(bound)):
                if i==0:
                    t[feature<=bound[i]] = i
                else:
                    t[(feature>bound[i-1])&(feature<=bound[i])] = i
            t[feature>bound[i]] = i+1
        else:
            t[feature<bound[0]] = 0
            for r, i in enumerate(bound):
                t[feature>=i] = r+1
        t = t.fillna(config['param']['miss_pad']).astype('int64').rename(config['param']['name'])
        if isinstance(config['param']['score'], (tuple, list)):
            t = t.replace({i:j for i,j in enumerate(config['param']['score'])})
            if config['param']['miss_score'] is not None:
                t = t.replace({config['param']['miss_pad']:config['param']['miss_score']})
        return t if mode else (t, config)


def numerical_combine(feature, function, mode=0, name=None, config=None):
    """feature combine transform.
    
    Args:
        feature: pd.DataFrame, sample feature.
        function: function or str, Function to use for aggregating the data. 
            If a function, must either work when passed a DataFrame or when passed to DataFrame.apply.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.        
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'function':function, 
                          'name':'combine_'+'_'.join([i.name for i in feature]) if name is None else name},
                  'type':'numerical_combine', 'variable':[i for i in feature.columns]}
    if mode==2:
        return config
    else:
        t = feature.agg(function, axis=1).rename(config['param']['name'])
        return t if mode else (t, config)


def numerical_cyclical(feature, mode=0, name=None, config=None):
    """feature cyclical transform.
    
    applies cyclical transformations to numerical variables, returning 2 new features per variable.
    according to:
        var_sin = sin(variable * (2. * pi / max_value))
        var_cos = cos(variable * (2. * pi / max_value))

    where max_value is the maximum value in the variable, and pi is 3.14…
    
    Args:
        feature: pd.Series, sample feature.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.        
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        if name is None:
            name = feature.name
        config = {'param':{'max_value':feature.max(), 'name':[f'{name}_sin', f'{name}_cos']},
                  'type':'numerical_cyclical', 'variable':feature.name}
    if mode==2:
        return config
    else:
        t = feature*(2.*np.pi/config['param']['max_value'])
        t = pd.concat([np.sin(t), np.cos(t)], axis=1)
        t.columns = config['param']['name']
        return t if mode else (t, config)


def numerical_math(feature, mode=0, method=['log'], log='e', c='auto', power=0.5, name=None, config=None):
    """feature math transform.
    
    using:
    - log
    - power
    - arcsin
    - reciprocal
    - BoxCox
    - YeoJohnson
    
    log:
        x = log(x+c)
        if c='auto', c = x.min().abs() + 1
        if x<0, x=0
        
        if log='e', x = log-e(x+c)
        if log='2', x = log-2(x+c)
        if log='10', x = log-10(x+c)
    
    power:
        x = x^power
        if power<1, x[x<=0] = 1
        
    arcsin:
        x = arcsin(sqrt(x))
        x must be 0~1
    
    reciprocal:
        x = 1/x
        if x=0, x=0
    
    BoxCox:
        applies the BoxCox transformation to numerical variables.
        The Box-Cox transformation is defined as:
        - T(Y)=(Y exp(λ)−1)/λ if λ!=0
        - log(Y) otherwise
        where Y is the response variable and λ is the transformation parameter. 
        λ varies, typically from -5 to 5. 
        In the transformation, all values of λ are considered and the optimal value for a given variable is selected.
        `Box and Cox. “An Analysis of Transformations”. Read at a RESEARCH MEETING, 1964.`
    
    YeoJohnson:
        applies the Yeo-Johnson transformation to the numerical variables.
        `Yeo, In-Kwon and Johnson, Richard (2000). 
        A new family of power transformations to improve normality or symmetry. Biometrika, 87, 954-959.`
    
    Args:
        feature: pd.Series, sample feature.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        method: list, must be one or more of ['log', 'power', 'arcsin', 'reciprocal']
        log: Indicates if the natural or base 10 logarithm should be applied. Can take values ‘e’ or ‘10’, '2'.
        c: The constant C to add to the variable before the logarithm, i.e., log(x + C).
        power: The power (or exponent).
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    method_list = ['log', 'power', 'arcsin', 'reciprocal', 'BoxCox', 'YeoJohnson']
    if config is None:
        if isinstance(method, str):
            method = [method]
        config = {'param':{'method':method, 'log':log, 'c':c,
                          'negative_replace':negative_replace, 'power':power,
                          'name':feature.name if name is None else name},
                  'type':'numerical_math', 'variable':feature.name}
        for i in method:
            if i not in method_list:
                raise ValueError(f"`method` must be one of {method_list}.")
        if len(method)!=len(set(method)):
            raise ValueError("`method` must be unique.")
        if 'log' in method:
            if c=='auto':
                config['param']['c'] = feature.min().abs() + 1
        if 'BoxCox' in method:
            import scipy.stats.boxcox as boxcox
            _, config['param']['boxcox_lambda'] = boxcox(feature)
        if 'YeoJohnson' in method:
            import scipy.stats.yeojohnson as yeojohnson
            _, config['param']['yeojohnson_lambda'] = yeojohnson(feature)
            
    if mode==2:
        return config
    else:
        log = config['param']['log']
        power = config['param']['power']
        
        t = feature.copy().rename(config['param']['name'])
        for i in config['param']['method']:
            if i=='log':
                t = t+config['param']['c']
                temp = t<=0
                if log=='e':
                    t = np.log(t)
                elif log=='2':
                    t = np.log2(t)
                elif log=='10':
                    t = np.log10(t)
                t[temp] = 0
            elif i=='power':
                if power<1:
                    t[t<=0] = 1
                t = np.power(t, power)
            elif i=='arcsin':
                t = np.arcsin(np.sqrt(t))
            elif i=='reciprocal':
                temp = t==0
                t = 1/t
                t[temp] = 0
            elif i=='BoxCox':
                import scipy.stats.boxcox as boxcox
                t = boxcox(t, lmbda=config['param']['boxcox_lambda'])
            elif i=='YeoJohnson':
                import scipy.stats.yeojohnson as yeojohnson
                t = yeojohnson(t, lmbda=config['param']['yeojohnson_lambda'])
            else:
                raise ValueError(f"`method` must be one of {method_list}.")
    return t if mode else (t, config)


def numerical_padding(feature, mode=0, method='mean', name=None, config=None):
    """feature fillna method.
    
    Args:
        feature: pd.Series, sample feature.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        method: default 'mean', one of 'mean' or 'median' or float or int.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'method':method, 'name':feature.name if name is None else name},
                  'type':'numerical_padding', 'variable':feature.name}
        if method=='mean':
            config['param']['feature_scale'] = feature.mean()
        elif method=='median':
            config['param']['feature_scale'] = feature.median()
        else:
            config['param']['feature_scale'] = method
    if mode==2:
        return config
    else:
        scale = config['param']['feature_scale']
        t = feature.fillna(scale).rename(config['param']['name'])
        return t if mode else (t, config)


def numerical_outlier(feature, mode=0, method='norm', delta=0.9545, tail='right', name=None, config=None):
    """feature clip outlier.
    
    Caps maximum and/or minimum values of a variable at automatically determined values, and optionally adds indicators
    The extreme values beyond which an observation is considered an outlier are determined 
    
    using:
    - norm
    - a Gaussian approximation
    - the inter-quantile range proximity rule (IQR)
    - MAD-median rule (MAD)
    - percentiles

    norm limits:
        right tail: mean + ppf((1-delta)/2)* std
        left tail: mean - ppf(delta+(1-delta)/2)* std
    
    Gaussian limits(Delta is recommended to be 3):
        right tail: mean + 3* std
        left tail: mean - 3* std

    IQR limits(Delta is recommended to be 3):
        right tail: 75th quantile + 3* IQR
        left tail: 25th quantile - 3* IQR
    where IQR is the inter-quartile range: 75th quantile - 25th quantile.

    MAD limits(Delta is recommended to be 3):
        right tail: median + 3* MAD
        left tail: median - 3* MAD
    where MAD is the median absoulte deviation from the median.

    percentiles:
        right tail: 95th percentile
        left tail: 5th percentile

    You can select how far out to cap the maximum or minimum values with the parameter 'delta'.
    If capping_method='gaussian' delta gives the value to multiply the std.
    If capping_method='iqr' delta is the value to multiply the IQR.
    If capping_method='mad' delta is the value to multiply the MAD.
    If capping_method='quantiles', delta is the percentile on each tail that should be censored. 
        For example, if delta=0.05, the limits will be the 5th and 95th percentiles. 
        If delta=0.1, the limits will be the 10th and 90th percentiles.
    Args:
        feature: pd.Series, sample feature.
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.
        method: must be one of ['norm', 'gaussian', 'iqr', 'mad', 'quantiles']
        delta: float, default 0.9545
        tail: str, default 'right', one of ['left', 'right', 'both'], statistical distribution boundary.
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if config is None:
        config = {'param':{'method':method, 'delta':delta, 'tail':tail,
                           'name':feature.name if name is None else name},
                  'type':'numerical_outlier', 'variable':feature.name}
        if method =='norm':
            if not 0<delta<1:
                raise ValueError("`delta` must be 0<delta<1")
            from scipy.stats import norm
            config['param']['feature_scale'] = [feature.mean(), feature.std()]
        elif method=='gaussian':
            config['param']['feature_scale'] = [feature.mean(), feature.std()]
        elif method=='iqr':
            config['param']['feature_scale'] = [feature.quantile(0.25), feature.quantile(0.75)]
        elif method=='mad':
            config['param']['feature_scale'] = [feature.median(), (feature-feature.median()).abs().median()]
        elif method=='quantiles':
            if not 0<delta<0.5:
                raise ValueError("`delta` must be 0<delta<0.5")
            config['param']['feature_scale'] = [feature.quantile(delta), feature.quantile(1-delta)]
        else:
            raise ValueError("`method` must be one of ['norm', 'gaussian', 'iqr', 'mad', 'quantiles']")
    if mode==2:
        return config
    else:
        scale = config['param']['feature_scale']
        delta = config['param']['delta']
        if method =='norm':
            if config['param']['tail']=='both':
                clip = (scale[0]+norm.ppf((1-delta)/2)*scale[1], scale[0]+norm.ppf(delta+(1-delta)/2)*scale[1])
            elif config['param']['tail']=='right':
                clip = (feature.min(), scale[0]+norm.ppf(delta)*scale[1])
            else:
                clip = (scale[0]+norm.ppf(1-delta)*scale[1], feature.max())
        elif method in ['gaussian', 'mad']:
            if config['param']['tail']=='both':
                clip = (scale[0]-delta*scale[1], scale[0]+delta*scale[1])
            elif config['param']['tail']=='right':
                clip = (feature.min(), scale[0]+delta*scale[1])
            else:
                clip = (scale[0]-delta*scale[1], feature.max())
        elif method =='iqr':
            if config['param']['tail']=='both':
                clip = (scale[0]-delta*(scale[1]-scale[0]), scale[1]+delta*(scale[1]-scale[0]))
            elif config['param']['tail']=='right':
                clip = (feature.min(), scale[1]+delta*(scale[1]-scale[0]))
            else:
                clip = (scale[0]-delta*(scale[1]-scale[0]), feature.max())
        elif method =='quantiles':
            if config['param']['tail']=='both':
                clip = (scale[0], scale[1])
            elif config['param']['tail']=='right':
                clip = (feature.min(), scale[1])
            else:
                clip = (scale[0], feature.max())
        else:
            raise ValueError("`method` must be one of ['norm', 'gaussian', 'iqr', 'mad', 'quantiles']")
        t = feature.clip(clip[0], clip[1]).rename(config['param']['name'])
        return t if mode else (t, config)


def numerical_relative(feature, reference, function, mode=0, name=None, config=None):
    """feature combine transform.
    
    supported method:
        ['add', 'sub', 'mul', 'div', 'truediv', 'floordiv', 'mod', 'pow', 'max', 'min', 'mean']
    
    Args:
        feature: pd.Series or pd.DataFrame, sample feature.
        reference: pd.Series or pd.DataFrame, sample feature.
        function: str or str of list, 
            one of ['add', 'sub', 'mul', 'div', 'truediv', 'floordiv', 'mod', 'pow', 'max', 'min', 'mean']
        mode: if 0, output (transform feature, config); if 1, output transform feature; if 2, output config.        
        name: str, output feature name, if None, name is feature.name .
        config: dict, label parameters dict for this estimator. 
            if config is not None, only parameter `feature` and `mode` is invalid.
    Returns:
        Refer to params `mode` explanation.
    """
    if type(feature)==pd.core.series.Series:
        feature = feature.to_frame()
    if type(reference)==pd.core.series.Series:
        reference = reference.to_frame()
    if isinstance(function, str):
        function = [function]
    function_list = ['add', 'sub', 'mul', 'div', 'truediv', 'floordiv', 'mod', 'pow', 'max', 'min', 'mean']
    for i in function:
        assert i in function_list, f'`function` must be one of {function_list}.'
    
    if config is None:
        config = {'param':{'function':function, 'reference':reference.columns.tolist(),
                          'name':'relative' if name is None else name},
                  'type':'numerical_relative', 'variable':feature.columns.tolist()}
    if mode==2:
        return config
    else:
        t = []
        
        for i in feature.columns:
            for j in reference.columns:
                for k in function:
                    if k=='add':
                        temp = feature[i].add(reference[j])
                    elif k=='sub':
                        temp = feature[i].sub(reference[j])
                    elif k=='mul':
                        temp = feature[i].mul(reference[j])
                    elif k=='div':
                        temp = feature[i].div(reference[j])
                    elif k=='truediv':
                        temp = feature[i].truediv(reference[j])
                    elif k=='floordiv':
                        temp = feature[i].floordiv(reference[j])
                    elif k=='mod':
                        temp = feature[i].mod(reference[j])
                    elif k=='pow':
                        temp = feature[i].pow(reference[j])
                    elif k=='max':
                        temp = pd.concat([feature[i], reference[j]], axis=1).max(axis=1)
                    elif k=='min':
                        temp = pd.concat([feature[i], reference[j]], axis=1).min(axis=1)
                    elif k=='mean':
                        temp = pd.concat([feature[i], reference[j]], axis=1).mean(axis=1)
                    t.append(temp.rename(f"{config['param']['name']}_{i}_{k}_{j}"))
        if len(t)>1:
            t = pd.concat(t, axis=1)
        return t if mode else (t, config)





from matplotlib import patches


class Rectangle():
    def add_rectangle(self, name, xy, width, height, angle=0.0, **kwargs):
        """Make a rectangle plot.
        
        The rectangle extends from ``xy[0]`` to ``xy[0] + width`` in x-direction
        and from ``xy[1]`` to ``xy[1] + height`` in y-direction. ::

          :                +------------------+
          :                |                  |
          :              height               |
          :                |                  |
          :               (xy)---- width -----+

        Args:
            xy: (float, float) The anchor point.
            width : float, Rectangle width.
            height : float, Rectangle height.
            angle : float, Rotation in degrees anti-clockwise about *xy*.
            alpha: scalar or None
            animated: bool
            antialiased or aa: unknown
            capstyle: `.CapStyle` or {'butt', 'projecting', 'round'}
            clip_box: `.Bbox`
            clip_on: bool
            clip_path: Patch or (Path, Transform) or None
            color: color
            contains: unknown
            edgecolor or ec: color or None or 'auto'
            facecolor or fc: color or None
            fill: bool
            gid: str
            hatch: {'/', '\\', '|', '-', '+', 'x', 'o', 'O', '.', '*'}
            in_layout: bool
            joinstyle: `.JoinStyle` or {'miter', 'round', 'bevel'}
            label: object
            linestyle or ls: {'-', '--', '-.', ':', '', (offset, on-off-seq), ...}
            linewidth or lw: float or None
            picker: None or bool or float or callable
            rasterized: bool
            sketch_params: (scale: float, length: float, randomness: float)
            snap: bool or None
        """
        if 'color' not in kwargs:
            kwargs['color'] = self._params.color.pop(0)[1]
        kwargs['width'] = width
        kwargs['height'] = height
        kwargs['angle'] = angle
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['data'] = xy
        self._params.ydata[name]['transform'] = 'ax'
        self._params.ydata[name]['plotmode'] = 'rectangle'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_rectangle
        return self
    
    def _execute_plot_rectangle(self, fig, ax, i, j):
        poly = patches.Rectangle(j['data'], **j['kwargs'])
        if j['transform']=='ax':
            ax.add_patch(poly)
        else:
            fig.add_artist(poly)


class Boxplot():
    def add_boxplot(self, name, xdata, **kwargs):
        """Make a box plot.
        
        Args:
            name: data name.
            xdata: x-axis data.
            notch : bool, default: False
                Whether to draw a notched box plot (`True`), or a rectangular box
                plot (`False`).  The notches represent the confidence interval (CI)
                around the median.  The documentation for *bootstrap* describes how
                the locations of the notches are computed by default, but their
                locations may also be overridden by setting the *conf_intervals*
                parameter.

            sym : str, optional
                The default symbol for flier points.  An empty string ('') hides
                the fliers.  If `None`, then the fliers default to 'b+'.  More
                control is provided by the *flierprops* parameter.

            vert : bool, default: True
                If `True`, draws vertical boxes.
                If `False`, draw horizontal boxes.

            whis : float or (float, float), default: 1.5
                The position of the whiskers.

                If a float, the lower whisker is at the lowest datum above
                ``Q1 - whis*(Q3-Q1)``, and the upper whisker at the highest datum
                below ``Q3 + whis*(Q3-Q1)``, where Q1 and Q3 are the first and
                third quartiles.  The default value of ``whis = 1.5`` corresponds
                to Tukey's original definition of boxplots.

                If a pair of floats, they indicate the percentiles at which to
                draw the whiskers (e.g., (5, 95)).  In particular, setting this to
                (0, 100) results in whiskers covering the whole range of the data.

                In the edge case where ``Q1 == Q3``, *whis* is automatically set
                to (0, 100) (cover the whole range of the data) if *autorange* is
                True.

                Beyond the whiskers, data are considered outliers and are plotted
                as individual points.

            bootstrap : int, optional
                Specifies whether to bootstrap the confidence intervals
                around the median for notched boxplots. If *bootstrap* is
                None, no bootstrapping is performed, and notches are
                calculated using a Gaussian-based asymptotic approximation
                (see McGill, R., Tukey, J.W., and Larsen, W.A., 1978, and
                Kendall and Stuart, 1967). Otherwise, bootstrap specifies
                the number of times to bootstrap the median to determine its
                95% confidence intervals. Values between 1000 and 10000 are
                recommended.

            usermedians : 1D array-like, optional
                A 1D array-like of length ``len(x)``.  Each entry that is not
                `None` forces the value of the median for the corresponding
                dataset.  For entries that are `None`, the medians are computed
                by Matplotlib as normal.

            conf_intervals : array-like, optional
                A 2D array-like of shape ``(len(x), 2)``.  Each entry that is not
                None forces the location of the corresponding notch (which is
                only drawn if *notch* is `True`).  For entries that are `None`,
                the notches are computed by the method specified by the other
                parameters (e.g., *bootstrap*).

            positions : array-like, optional
                The positions of the boxes. The ticks and limits are
                automatically set to match the positions. Defaults to
                ``range(1, N+1)`` where N is the number of boxes to be drawn.

            widths : float or array-like
                The widths of the boxes.  The default is 0.5, or ``0.15*(distance
                between extreme positions)``, if that is smaller.

            patch_artist : bool, default: False
                If `False` produces boxes with the Line2D artist. Otherwise,
                boxes and drawn with Patch artists.

            labels : sequence, optional
                Labels for each dataset (one per dataset).

            manage_ticks : bool, default: True
                If True, the tick locations and labels will be adjusted to match
                the boxplot positions.

            autorange : bool, default: False
                When `True` and the data are distributed such that the 25th and
                75th percentiles are equal, *whis* is set to (0, 100) such
                that the whisker ends are at the minimum and maximum of the data.

            meanline : bool, default: False
                If `True` (and *showmeans* is `True`), will try to render the
                mean as a line spanning the full width of the box according to
                *meanprops* (see below).  Not recommended if *shownotches* is also
                True.  Otherwise, means will be shown as points.

            zorder : float, default: ``Line2D.zorder = 2``
                The zorder of the boxplot.
        """
        self._params.ydata[name]['kwargs'] = kwargs
        self._params.ydata[name]['xdata'] = xdata
        self._params.ydata[name]['plotmode'] = 'boxplot'
        self._params.ydata[name]['plotfunc'] = self._execute_plot_boxplot
        return self
    
    def _execute_plot_boxplot(self, fig, ax, i, j):
        axplot = ax.boxplot(j['xdata'], **j['kwargs'])


import os
import poplib
import smtplib
from base64 import b64encode
from email.encoders import encode_base64
from email.header import Header
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import pandas as pd

from linora.utils._config import Config
from linora.utils._logger import Logger
from linora.utils._email_utils import parse_mail, parse_headers

__all__ = ['EMail']


poplib._MAXLINE = 4096
mail_personal = {
    '163.com': {
        'smtp_host': 'smtp.163.com',
        'smtp_port': 994,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.163.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
        'imap_host': 'imap.163.com',
        'imap_port': 993,
        'imap_ssl': True,
        'imap_tls': False
    },
    '126.com': {
        'smtp_host': 'smtp.126.com',
        'smtp_port': 994,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.126.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
        'imap_host': 'imap.126.com',
        'imap_port': 993,
        'imap_ssl': True,
        'imap_tls': False
    },
    'yeah.net': {
        'smtp_host': 'smtp.yeah.net',
        'smtp_port': 994,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.yeah.net',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
        'imap_host': 'imap.yeah.net',
        'imap_port': 993,
        'imap_ssl': True,
        'imap_tls': False
    },
    'qq.com': {
        'smtp_host': 'smtp.qq.com',
        'smtp_port': 465,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.qq.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
    },
    'gmail.com': {
        'smtp_host': 'smtp.gmail.com',
        'smtp_port': 465,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.gmail.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
    },
    'sina.com': {
        'smtp_host': 'smtp.sina.com',
        'smtp_port': 465,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.sina.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
    },
    'outlook.com': {
        'smtp_host': 'smtp-mail.outlook.com',
        'smtp_port': 587,
        'smtp_ssl': False,
        'smtp_tls': True,
        'pop_host': 'pop.outlook.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
    },
    'hotmail.com': {
        'smtp_host': 'smtp.office365.com',
        'smtp_port': 587,
        'smtp_ssl': False,
        'smtp_tls': True,
        'pop_host': 'outlook.office365.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
    },
}

mail_enterprise = {
    'qq': {
        'smtp_host': 'smtp.exmail.qq.com',
        'smtp_port': 465,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.exmail.qq.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False
    },
    'ali': {
        'smtp_host': 'smtp.mxhichina.com',
        'smtp_port': 465,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop3.mxhichina.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False
    },
    '163': {
        'smtp_host': 'smtp.qiye.163.com',
        'smtp_port': 994,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.qiye.163.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False
    },
    'google': {
        'smtp_host': 'smtp.gmail.com',
        'smtp_port': 465,
        'smtp_ssl': True,
        'smtp_tls': False,
        'pop_host': 'pop.gmail.com',
        'pop_port': 995,
        'pop_ssl': True,
        'pop_tls': False,
    },
}

mail_default = {
    'smtp_host': 'smtp.',
    'smtp_port': 465,
    'smtp_ssl': True,
    'smtp_tls': False,
    'pop_host': 'pop.',
    'pop_port': 995,
    'pop_ssl': True,
    'pop_tls': False,
    'imap_host': 'imap.',
    'imap_port': 993,
    'imap_ssl': True,
    'imap_tls': False
}

class EMail():
    def __init__(self, username, password, smtp_host=None, smtp_port=None, smtp_ssl=None, smtp_tls=None,
                 pop_host=None, pop_port=None, pop_ssl=None, pop_tls=None, enterprise=None):
        """ Return MailServer instance, it implements all SMTP and POP functions.
        
        The module is inspired by zmail.
        
        Args:
            username: email name.
            password: email password.
            smtp_host: email smtp host.
            smtp_port: email smtp port.
            smtp_ssl: email smtp ssl.
            smtp_tls: email smtp tls.
            pop_host: email pop host.
            pop_port: email pop port.
            pop_ssl: email pop ssl.
            pop_tls: email pop tls.
            enterprise: Shortcut for use enterprise mail,if specified, enterprise mail configs will replace all inner auto-generate configs. support ['qq', 'ali', '163', 'google']
        """
        warn = Logger(name='')
        warn.info(f"la.utils.EMail has been deprecated and will be deleted in version 2.1.0. Please use la.server.EMail")
        self._params = Config()
        self._params.mail_personal = mail_personal
        self._params.mail_enterprise = mail_enterprise
        self._params.mail_default = mail_default
        self._params.username = username
        self._params.password = password
        
        user_define_config = {
            'smtp_host': smtp_host,
            'smtp_port': smtp_port,
            'smtp_ssl': smtp_ssl,
            'smtp_tls': smtp_tls,
            'pop_host': pop_host,
            'pop_port': pop_port,
            'pop_ssl': pop_ssl,
            'pop_tls': pop_tls}
        mail_config = self._check_supported_server(username, enterprise)
        mail_config.update({k: v for k, v in mail_config.items() if v is not None})
        self._params.mail_config = {k: v for k, v in mail_config.items() if 'imap' not in k}
        self._params.mime = None
        self._params.timeout = 60
        
    def send_mail(self, recipients, mail, cc=None, timeout=None, auto_add_from=True, auto_add_to=True):
        """"Send email.
        
        Args:
            recipients: recipient user email, can either be str or a list of str.
            mail: dict, mail content.
            cc: Mail cc object.
            timeout: if is not None, it will replace server's timeout.
            auto_add_from: If set to True, when the key ' 'from' not in mail, the default 'from' will automatically added to mail.
            auto_add_to: If set to True, when the key 'to' not in mail, the default 'to' will automatically added to mail.
        """
        if timeout is not None:
            self._params.timeout = timeout
        if auto_add_from and mail.get('From') is None:
            if self._params.mime is None:
                self._make_mime(mail)
            self._params.mime['From'] = self._make_address_header([self._params.username])
        
        recipients = recipients if  isinstance(recipients, list) else [recipients]
        if auto_add_to and mail.get('To') is None:
            if self._params.mime is None:
                self._make_mime(mail)
            self._params.mime['To'] = self._make_address_header(recipients)
            
        if cc is not None:
            cc = cc if isinstance(cc, list) else [cc]
            recipients.extend(cc)
            if self._params.mime is None:
                self._make_mime(mail)
            self._params.mime['Cc'] = self._make_address_header(cc)
        
        recipients = [i[1] if isinstance(i, tuple) else i for i in recipients]
        self._login()
        self._params.server.timeout = self._params.timeout
        self._params.server.sendmail(self._params.username, recipients, self._params.mime.as_string())
        self._logout()
        self._params.mime = None
        return True
    
    def delete(self, which):
        """Delete mail.
        
        Args:
            which: int, which is a int number that represent mail's position in mailbox.
                The which must between 1 and message count(return from Mail.stat())
        """
        self._login(send=False)
        self._params.server.dele(which)
        self._logout(send=False)
        return True

    def stat(self):
        """Get mailbox status.
        The result is a tuple of 2 integers: (message count, mailbox size).
        """
        self._login(send=False)
        status = self._params.server.stat()
        self._logout(send=False)
        return status

    def get_mail(self, which=None, subject=None, sender=None, start_time=None, end_time=None, 
                  start_index=None, end_index=None):
        """Get a list of mails from mailbox.
        
        Args:
            which:int or list, which is a int number that represent mail's position in mailbox.
                The which must between 1 and message count(return from Mail.stat())
            subject: email subject
            sender: email sender
            start_time: email start time
            end_time: email end time
            start_index: email start index
            end_index: email end index
        """
        if start_time is not None:
            start_time = pd.to_datetime(start_time)
        if end_time is not None:
            end_time = pd.to_datetime(end_time)
        if which is not None:
            if not isinstance(which, (list, tuple)):
                which = [which]
        elif subject or sender or start_time or end_time or start_index or end_index:
            headers = self.get_headers(start_index, end_index)

            which = []
            for header in headers:
                mail_subject = header.get('Subject')
                mail_sender = header.get('From')
                mail_date = header.get('date')
                ok = True
                if subject is not None:
                    if mail_subject is None or subject not in mail_subject:
                        ok = False
                if sender is not None:
                    if mail_sender is None or sender not in mail_sender:
                        ok = False
                if start_time is not None:
                    if mail_date is None or start_time > mail_date:
                        ok = False
                if end_time is not None:
                    if mail_date is None or end_time < mail_date:
                        ok = False
                if ok:
                    which.append(header['id'])
            which.sort()
        else:
            which = [self.stat()[0]]
        self._login(send=False)
        mail = [parse_mail(self._params.server.retr(i)[1], i) for i in which]
        self._logout(send=False)
        return mail

    def get_headers(self, start_index=None, end_index=None):
        """Get mails headers.
        
        Args:
            start_index: email start index
            end_index: email end index
        """
        self._login(send=False)
        end = self._params.server.stat()[0]
        if start_index is None:
            start_index = 1
        if end_index is None:
            end_index = end
        index = range(min(start_index, end_index), max(start_index, end_index)+1)
        header = [self._params.server.top(i, 0)[1] for i in index]
        headers = []
        for i, mail_header in enumerate(header):
            _, _headers, *__ = parse_headers(mail_header)
            _headers.update(id=index[i])
            try:
                headers.append({j:_headers[j] for j in {'date', 'Subject','From', 'id'}})
            except:
                print(_headers)
        self._logout(send=False)
        return headers

    def check_available(self):
        """check email connection status."""
        try:
            self._login()
            self._logout()
            check = {'stmp server':True, 'stmp info':'ok'}
        except Exception as e:
            check = {'stmp server':False, 'stmp info':str(e)}
        try:
            self._login(send=False)
            self._logout(send=False)
            check.update({'pop server':True, 'pop info':'ok'})
        except Exception as e:
            check.update({'pop server': False, 'pop info':str(e)})
        return check
        
    def _check_supported_server(self, mail_address, config=None):
        provider = mail_address.split('@')[1]

        if config is not None:
            if config in self._params.mail_enterprise:
                return self._params.mail_enterprise[config]
            else:
                raise ValueError(f'Only supported enterprise server [{str(list(self._params.mail_enterprise))[1:-1]}].')

        if provider in self._params.mail_personal:
            return self._params.mail_personal[provider]
        else:
            config = self._params.mail_default.copy()
            config['smtp_host'] += provider
            config['pop_host'] += provider
            config['imap_host'] += provider
            return config
        
    def _login(self, send=True):
        if send:
            if self._params.mail_config['smtp_ssl']:
                self._params.server = smtplib.SMTP_SSL(self._params.mail_config['smtp_host'], self._params.mail_config['smtp_port'], 'email.local', timeout=self._params.timeout)
            else:
                self._params.server = smtplib.SMTP(self._params.mail_config['smtp_host'], self._params.mail_config['smtp_port'], 'email.local', timeout=self._params.timeout)
            if self._params.mail_config['smtp_tls']:
                self._params.server.ehlo()
                self._params.server.starttls()
                self._params.server.ehlo()
            self._params.server.login(self._params.username, self._params.password)
        else:
            if self._params.mail_config['pop_ssl']:
                self._params.server = poplib.POP3_SSL(self._params.mail_config['pop_host'], self._params.mail_config['pop_port'], timeout=self._params.timeout)
            else:
                self._params.server = poplib.POP3(self._params.mail_config['pop_host'], self._params.mail_config['pop_port'], timeout=self._params.timeout)
            if self._params.mail_config['pop_tls']:
                self._params.server.stls()
            self._params.server.user(self._params.username)
            self._params.server.pass_(self._params.password)
                
    def _logout(self, send=True):
        if send:
            try:
                code, message = self._params.server.docmd("QUIT")
                if code != 221:
                    raise smtplib.SMTPResponseException(code, message)
            except smtplib.SMTPServerDisconnected:
                pass
            finally:
                self._params.server.close()
        else:
            self._params.server.quit()
        
    def _make_mime(self, mail):
        mime = MIMEMultipart(boundary=None)
        for k, v in mail.items():
            _k = k.lower()
            if _k in ('subject', 'from'):
                if isinstance(v, str):
                    mime[_k.capitalize()] = v
                else:
                    raise InvalidArguments('{} can only be str! Got {} instead.'.format(_k.capitalize(), type(v)))
#             elif _k == 'to':
                
#                 if not all([(i in mail) for i in
#                     ('from', 'to', 'subject', 'raw_headers', 'charsets', 'headers',
#                      'date', 'id', 'raw', 'attachments', 'content_text', 'content_html')]):
#                     warnings.warn("Header 'to' is invalid and unused,if you want to add address name "
#                                   "use tuple (address,name) instead.", category=DeprecationWarning, stacklevel=4)
#             elif _k in ('attachments', 'content_text', 'content_html', 'headers'):
#                 pass
#             else:
#                 if not self._is_resend_mail():
                    # Remove resend warnings.
#                     warnings.warn("Header '{}' is invalid and unused,if you want to add extra headers "
#                                   "use 'headers' instead.".format(str(_k)), category=DeprecationWarning, stacklevel=4)

        # Set extra headers.
        if mail.get('headers') and isinstance(mail['headers'], dict):
            for k, v in mail['headers'].items():
                mime[k] = v

        # Set HTML content.
        if mail.get('content_html') is not None:
            html = mail['content_html'] if isinstance(mail['content_html'], list) else [mail['content_html']]
            for i in html:
                mime.attach(MIMEText('{}'.format(i), 'html', 'utf-8'))

        # Set TEXT content.
        if mail.get('content_text') is not None:
            text = mail['content_text'] if isinstance(mail['content_text'], list) else [mail['content_text']]
            for i in text:
                mime.attach(MIMEText('{}'.format(i), 'plain', 'utf-8'))

        # Set attachments.
        if mail.get('attachments'):
            attachments = mail['attachments'] if isinstance(mail['attachments'], list) else [mail['attachments']]
            for attachment in attachments:
                if isinstance(attachment, str):
                    if not os.path.exists(attachment):
                        raise FileExistsError(f"The file {attachment} doesn't exist.")
                    name = os.path.split(attachment)[1]
                    encoded_name = Header(name).encode()
                    with open(attachment, 'rb') as f:
                        part = MIMEBase('application', 'octet-stream')
                        part.set_payload(f.read())
                        part['Content-Disposition'] = f'attachment;filename="{encoded_name}"'
                        encode_base64(part)
                    mime.attach(part)
                elif isinstance(attachment, tuple):
                    name, raw = attachment
                    part = MIMEBase('application', 'octet-stream')
                    part.set_payload(raw)
                    part['Content-Disposition'] = f'attachment;filename="{name}"'
                    encode_base64(part)
                    mime.attach(part)
                else:
                    raise InvalidArguments(f'Attachments excepted str or tuple got {type(attachment)} instead.')
        self._params.mime = mime
        
    def _make_address_header(self, address_list):
        """Used for make 'To' 'Cc' 'From' header."""
        res = []
        for address in address_list:
            if isinstance(address, tuple):
                assert len(address) == 2, 'Only two arguments!'
                name, rel_address = address
                name = '' if not name else f"=?utf-8?b?{b64encode(name.encode('utf-8')).decode('ascii')}?="
                res.append(f'{name} <{rel_address}>')
            elif isinstance(address, str):
                res.append(f'<{address}>')
            else:
                raise InvalidArguments(f'Email address can only be tuple or str.Get {type(address)} instead.')
        return ', '.join(res)
    
    


import os
import sys
import json
import time
import pickle
from collections import Counter
from multiprocessing import cpu_count

import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedKFold, train_test_split
from linora.param_search._HyperParameters import HyperParametersRandom
from linora.param_search._config import __xgboost_version__


def RandomSearch(feature, label, group, metrics, iter_num=1000, scoring=0.5, cv=5, cv_num=3,
                 metrics_min=True, speedy=True, speedy_param=(20000, 0.3), gpu=False,
                 save_model_dir=None
                ):
    """XGBRanker model params search use RandomSearch method.
    
    Args:
        feature: pandas dataframe, model's feature.
        label: pandas series, model's label.
        group: label group.
        metrics: model metrics function.
        scoring: metrics error opt base line value.
        cv: cross validation fold.
        cv_num: minimum cross validation fold.
        metrics_min: metrics value whether the smaller the better.
        speedy: whether use speedy method.
        speedy_param: if use speedy method, test_size will be set, 
                      test_size = 1-round(min(speedy_param[0], feature.shape[0]*speedy_param[1])/feature.shape[0], 2).
        gpu: whether use gpu.
        save_model_dir: save model folder.
    Returns:
        a best XGBRanker model params dict.
    Raises:
        params error.
    """
    import xgboost as xgb
    assert xgb.__version__>=__xgboost_version__, f'xgboost version should be >={__xgboost_version__}.'
    start = time.time()
    if gpu:
        raise "XGBRanker is not supported currently."
    best_params={}
    if speedy:
        test_size = 1-round(min(speedy_param[0], feature.shape[0]*speedy_param[1])/feature.shape[0], 2)
    tree_method = ['gpu_hist'] if gpu else ['auto', 'exact', 'approx', 'hist']
    n_job = 1 if gpu else int(np.ceil(cpu_count()*0.9))
    weight_dict = Counter(label)
    if len(weight_dict)==2:
        weight = int(np.ceil(weight_dict[min(weight_dict)]/weight_dict[max(weight_dict)]))
    else:
        weight_dict = {j:i for i,j in weight_dict.items()}
        weight = int(np.ceil(weight_dict[max(weight_dict)]/weight_dict[min(weight_dict)]))
    
    hp = HyperParametersRandom()
    hp.Float('learning_rate', 0.01, 0.1)
    hp.Int('n_estimators', 100, 850)
    hp.Choice('max_depth', [3, 4, 5, 6, 7])
    hp.Choice('min_child_weight', [1, 2, 3, 4, 5, 6, 7])
    hp.Choice('max_delta_step', [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    hp.Choice('reg_alpha', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
    hp.Choice('reg_lambda', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
    hp.Choice('subsample', [0.5, 0.6, 0.7, 0.8, 0.9, 1. ])
    hp.Choice('colsample_bytree', [0.5, 0.6, 0.7, 0.8, 0.9, 1. ])
    hp.Choice('colsample_bylevel', [0.5, 0.6, 0.7, 0.8, 0.9, 1. ])
    hp.Choice('colsample_bynode', [0.5, 0.6, 0.7, 0.8, 0.9, 1. ])
    hp.Choice('gamma', np.concatenate([np.linspace(0, 1, 101), np.linspace(2, 100, 99)]).round(2))
    hp.Choice('scale_pos_weight', [1, weight])
    hp.Choice('n_jobs', [n_job])
    hp.Choice('random_state', [27])
    hp.Choice('objective', ['rank:pairwise'])
    hp.Choice('booster', ['gbtree'])
    hp.Choice('tree_method', tree_method)
    hp.Choice('importance_type', ["gain", "weight", "cover", "total_gain", "total_cover"])
    
    for i in range(1, iter_num+1):
        hp.update()
        model = xgb.XGBRanker(**hp.params)
        score = []
        if speedy:
            for _ in range(cv_num):
                X_train, X_test, y_train, y_test, g_train, g_test = train_test_split(feature, label, group, 
                                                                                     test_size=test_size, stratify=label, 
                                                                                     random_state=np.random.choice(range(100), 1)[0])
                model.fit(X_train, y_train, g_train)
                cv_pred = model.predict(X_test)
                score.append(metrics(y_test.values, cv_pred))
        else:
            skf = StratifiedKFold(n_splits=cv, shuffle=True, random_state=np.random.choice(range(100), 1)[0])
            for n, (train_index, test_index) in enumerate(skf.split(feature, label)):
                if n == cv_num:
                    break
                model.fit(feature.loc[train_index], label[train_index], group[train_index])
                cv_pred = model.predict(feature.loc[test_index])
                score.append(metrics(label[test_index].values, cv_pred))
        cv_score = round(np.mean(score), 4)
        if metrics_min:
            if cv_score<scoring:
                scoring = cv_score
                best_params = params.copy()
                if save_model_dir is not None:
                    pickle.dump(model, open(os.path.join(save_model_dir, "xgb_model.pkl"), "wb"))
                    with open(os.path.join(save_model_dir, "xgb_params.json"),'w') as f:
                        json.dump(best_params, f)
        else:
            if cv_score>scoring:
                scoring = cv_score
                best_params = params.copy()
                if save_model_dir is not None:
                    pickle.dump(model, open(os.path.join(save_model_dir, "xgb_model.pkl"), "wb"))
                    with open(os.path.join(save_model_dir, "xgb_params.json"),'w') as f:
                        json.dump(best_params, f)
        sys.stdout.write("XGBRanker random search percent: {}%, run time {} min, best score: {}, best param：{}\r".format(
            round(i/iter_num*100,2), divmod((time.time()-start),60)[0], scoring, best_params))
        sys.stdout.flush()
    print("XGBRanker param finetuning with random search run time: %d min %.2f s" % divmod((time.time() - start), 60))
    return best_params



import random

import numpy as np

__all__ = ['skipgrams', 'sequence_pad', 'make_sampling_table']


def skipgrams(sequence, vocabulary_size,
              window_size=4, negative_samples=1., shuffle=True,
              categorical=False, sampling_table=None, seed=None):
    """Generates skipgram word pairs.
    
    This function transforms a sequence of word indexes (list of integers)
    into tuples of words of the form:
    - (word, word in the same window), with label 1 (positive samples).
    - (word, random word from the vocabulary), with label 0 (negative samples).
    Read more about Skipgram in this gnomic paper by Mikolov et al.:
    [Efficient Estimation of Word Representations in
    Vector Space](http://arxiv.org/pdf/1301.3781v3.pdf)
    
    Args:
        sequence: A word sequence (sentence), encoded as a list
            of word indices (integers). If using a `sampling_table`,
            word indices are expected to match the rank
            of the words in a reference dataset (e.g. 10 would encode
            the 10-th most frequently occurring token).
            Note that index 0 is expected to be a non-word and will be skipped.
        vocabulary_size: Int, maximum possible word index + 1
        window_size: Int, size of sampling windows (technically half-window).
            The window of a word `w_i` will be
            `[i - window_size, i + window_size+1]`.
        negative_samples: Float >= 0. 0 for no negative (i.e. random) samples.
            1 for same number as positive samples.
        shuffle: Whether to shuffle the word couples before returning them.
        categorical: bool. if False, labels will be
            integers (eg. `[0, 1, 1 .. ]`),
            if `True`, labels will be categorical, e.g.
            `[[1,0],[0,1],[0,1] .. ]`.
        sampling_table: 1D array of size `vocabulary_size` where the entry i
            encodes the probability to sample a word of rank i.
        seed: Random seed.
    Returns:
        couples, labels: where `couples` are int pairs and
            `labels` are either 0 or 1.
    Note
        By convention, index 0 in the vocabulary is
        a non-word and will be skipped.
    """
    if sampling_table is None:
        sampling_table = [1]*vocabulary_size
    if not isinstance(sequence[0], list):
        sequence = [sequence]
    couples = []
    labels = []
    for seq in sequence:
        t = [[i, wi, max(0, i-window_size), min(len(seq), i+window_size+1)] 
                   for i, wi in enumerate(seq) if wi and sampling_table[wi]>random.random()]
        couples += [[wi, seq[j]] for (i, wi, m, n) in t for j in range(m, n) if j!=i and seq[j]]
        labels += ([[0, 1]] if categorical else [1])*len(couples) 
    
    if negative_samples > 0:
        num = int(len(labels) * negative_samples)
        words = [c[0] for c in couples]
        random.shuffle(words)
        couples += [[words[i%len(words)], random.randint(1, vocabulary_size-1)] for i in range(num)]
        labels += ([[1, 0]] if categorical else [0])*num

    if shuffle:
        if seed is None:
            seed = random.randint(0, 10e6)
        random.seed(seed)
        random.shuffle(couples)
        random.seed(seed)
        random.shuffle(labels)
    return couples, labels


def sequence_pad(sequence, maxlen, dtype='int32', padding='pre', truncating='pre', value=0):
    """Pads sequences to the same length.
    
    Args:
        sequences: pd.Series or np.array or List of lists, where each element is a sequence.
        maxlen: Int, maximum length of all sequences.
        dtype: Type of the output sequences.
               To pad sequences with variable length strings, you can use `object`.
        padding: String, 'pre' or 'post':
                 pad either before or after each sequence.
        truncating: String, 'pre' or 'post':
                    remove values from sequences larger than
                    `maxlen`, either at the beginning or at the end of the sequences.
        value: Float or String, padding value.
    Returns:
        List of lists with shape `(len(sequences), maxlen)`
    """
    if not isinstance(sequence[0], list):
        sequence = [sequence]
    if padding=='post' and truncating=='post':
        t = [i[:maxlen] if len(i)>maxlen else i+[value]*(maxlen-len(i)) for i in sequence]
    elif padding=='post' and truncating=='pre':
        t = [i[-maxlen:] if len(i)>maxlen else i+[value]*(maxlen-len(i)) for i in sequence]
    elif padding=='pre' and truncating=='post':
        t = [i[:maxlen] if len(i)>maxlen else [value]*(maxlen-len(i))+i for i in sequence]
    elif padding=='pre' and truncating=='pre':
        t = [i[-maxlen:] if len(i)>maxlen else [value]*(maxlen-len(i))+i for i in sequence]
    else:
        raise ValueError('Padding type "%s" not understood or Truncating type "%s" not understood' % (padding, truncating))
    return t


def make_sampling_table(size, sampling_factor=1e-5):
    """Generates a word rank-based probabilistic sampling table.
    
    Used for generating the `sampling_table` argument for `skipgrams`.
    `sampling_table[i]` is the probability of sampling
    the word i-th most common word in a dataset
    (more common words should be sampled less frequently, for balance).
    The sampling probabilities are generated according
    to the sampling distribution used in word2vec:
    ```
    p(word) = (min(1, sqrt(word_frequency / sampling_factor) /
        (word_frequency / sampling_factor)))
    ```
    We assume that the word frequencies follow Zipf's law (s=1) to derive
    a numerical approximation of frequency(rank):
    `frequency(rank) ~ 1/(rank * (log(rank) + gamma) + 1/2 - 1/(12*rank))`
    where `gamma` is the Euler-Mascheroni constant.
    
    Args:
        size: Int, number of possible words to sample.
        sampling_factor: The sampling factor in the word2vec formula.
    Returns:
        A 1D Numpy array of length `size` where the ith entry
        is the probability that a word of rank i should be sampled.
    """
    gamma = 0.577
    rank = np.arange(size)
    rank[0] = 1
    inv_fq = rank * (np.log(rank) + gamma) + 0.5 - 1. / (12. * rank)
    f = sampling_factor * inv_fq
    return np.minimum(1., f / np.sqrt(f))



from itertools import chain
from collections import Counter

__all__ = ['sequence_preprocess', 'word_count', 'word_low_freq', 'word_high_freq', 'filter_word',
           'filter_punctuation', 'ngrams_iterator']


def sequence_preprocess(sequence):
    """Sequence preprocess, keep only Chinese.
    
    Args:
        sequence: pd.Series or np.array or list, sample feature value.
    Returns:
        a only Chinese list.
    """
    if isinstance(sequence, str):
        return ''.join([i for i in sequence.replace(' ', '') if i>= u'\u4e00' and i<= u'\u9fa5'])
    return [''.join([i for i in s.replace(' ', '') if i>= u'\u4e00' and i<= u'\u9fa5']) for s in sequence]


def word_count(sequence):
    """Sequence word count.
    
    Args:
        sequence: pd.Series or np.array or list of Lists, sample feature value.
    Returns:
        a dict with word count.
    """
    return Counter(chain.from_iterable(sequence))


def word_low_freq(word_count_dict, threshold=3):
    """Filter low frequency words.
    
    Args:
        word_count_dict: a dict with word count.
        threshold: filter low frequency words threshold.
    Returns:
        a list filter low frequency words.
    """
    return [i for i,j in word_count_dict.items() if j<=threshold]


def word_high_freq(word_count_dict, threshold):
    """Filter high frequency words.
    
    Args:
        word_count_dict: a dict with word count.
        threshold: filter high frequency words threshold.
    Returns:
        a list filter high frequency words.
    """
    return [i for i,j in word_count_dict.items() if j>=threshold]


def filter_word(sequence, filter_word_list):
    """Sequence filter words with a filter word list.
    
    Args:
        sequence: pd.Series or np.array or list of Lists, sample feature value.
        filter_word_list: a filter word list.
    Returns:
        a list of Lists.
    """
    if isinstance(sequence[0], str):
        return [x for x in sequence if x not in filter_word_list]
    return [[x for x in i if x not in filter_word_list] for i in sequence]


def filter_punctuation(sequence, punctuation=None):
    """Sequence preprocess, filter punctuation.
    
    Args:
        sequence: pd.Series or np.array or list, sample feature value.
        punctuation: str, filter punctuation.
    Returns:
        a list of Lists.
    """
    punc = '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~（），。’‘”“！？》《·、】【；：' if punctuation is None else punctuation
    table = str.maketrans('', '', punc)
    if isinstance(sequence, str):
        return sequence.translate(table)
    return [s.translate(table) for s in sequence]


def ngrams_iterator(token_list, ngrams, join=' '):
    """Return an iterator that yields the given tokens and their ngrams.

    Args:
        token_list: A list of tokens
        ngrams: the number of ngrams.
    """
    for x in token_list:
        yield x
    for n in range(2, ngrams + 1):
        for x in zip(*[token_list[i:] for i in range(n)]):
            yield join.join(x)


import numpy as np
import pandas as pd

from linora.metrics._utils import _sample_weight
from linora.metrics._regression import mean_squared_error

__all__ = ['accuracy_binary', 'accuracy_categorical', 'recall', 'precision', 'confusion_matrix',
           'fbeta_score', 'f1_score', 'auc_roc', 'auc_pr', 'crossentropy_binary', 
           'crossentropy_categorical', 'ks', 'gini', 'psi', 'fmi', 'report_binary',
           'accuracy_categorical_top_k', 'iou_binary', 'iou_categorical',
           'precision_on_recall', 'recall_on_precision', 
           'specificity_on_sensitivity', 'sensitivity_on_specificity', 'best_prob', 'matthews_score'
          ]


def classified_func(y_true, y_pred, prob=0.5, pos_label=1):
    t = pd.DataFrame({'prob':y_pred, 'label':y_true})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    if t.prob.nunique()!=2:
        label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
        t['label'] = t.label.replace(label_dict)
        t.loc[t.prob>=prob, 'prob'] = 1
        t.loc[t.prob<prob, 'prob'] = 0
    return t


def accuracy_binary(y_true, y_pred, sample_weight=None, prob=0.5, pos_label=1):
    """Calculates how often predictions match binary labels.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        the fraction of correctly classified samples (float).
    """
    sample_weight = _sample_weight(y_true, sample_weight)
    t = classified_func(y_true, y_pred, prob=prob, pos_label=pos_label)
    return ((t.label==t.prob)*sample_weight).mean()


def accuracy_categorical(y_true, y_pred, sample_weight=None):
    """Calculates how often predictions match categorical labels.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
    Returns:
        the fraction of correctly classified samples (float).
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    if y_true.ndim!=1:
        y_true = np.argmax(y_true, axis=-1)
    if y_pred.ndim!=1:
        y_pred = np.argmax(y_pred, axis=-1)
    sample_weight = _sample_weight(y_true, sample_weight)
    return ((pd.Series(y_true)==pd.Series(y_pred))*sample_weight).mean()


def matthews_score(y_true, y_pred, sample_weight=None, prob=0.5, pos_label=1):
    """Compute the Matthews correlation coefficient (MCC).

    The Matthews correlation coefficient is used in machine learning as a measure of 
    the quality of binary and multiclass classifications. 
    It takes into account true and false positives and negatives and is generally 
    regarded as a balanced measure which can be used even if the classes are of very different sizes. 
    The MCC is in essence a correlation coefficient value between -1 and +1. 
    A coefficient of +1 represents a perfect prediction, 
    0 an average random prediction and -1 an inverse prediction. 
    The statistic is also known as the phi coefficient.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        The Matthews correlation coefficient (+1 represents a perfect prediction, 0 an average random prediction and -1 and inverse prediction).
    """
    sample_weight = _sample_weight(y_true, sample_weight)
    t = classified_func(y_true, y_pred, prob=prob, pos_label=pos_label)
    t['weight'] = sample_weight
    tp = (((t.label==pos_label)&(t.prob==pos_label))*t.weight).sum()
    fp = (((t.label==pos_label)&(t.prob!=pos_label))*t.weight).sum()
    fn = (((t.label!=pos_label)&(t.prob==pos_label))*t.weight).sum()
    tn = (((t.label!=pos_label)&(t.prob!=pos_label))*t.weight).sum()
    return (tp*tn-fp*fn)/np.sqrt((tp+fp)*(tp+fn)*(tn+fp)*(tn+fn))


def iou_binary(y_true, y_pred, sample_weight=None, prob=0.5, pos_label=1):
    """Computes the Intersection-Over-Union metric for class 0 and/or 1.
    
    iou = true_positives / (true_positives + false_positives + false_negatives)
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        the fraction of correctly classified samples (float).
    """
    sample_weight = _sample_weight(y_true, sample_weight)
    t = classified_func(y_true, y_pred, prob=prob, pos_label=pos_label)
    t['weight'] = sample_weight
    tp = (((t.label==pos_label)&(t.prob==pos_label))*t.weight).sum()
    fp = (((t.label==pos_label)&(t.prob!=pos_label))*t.weight).sum()
    fn = (((t.label!=pos_label)&(t.prob==pos_label))*t.weight).sum()
    return tp/(tp+fp+fn)


def iou_categorical(y_true, y_pred, sample_weight=None, target_class_ids=None):
    """Computes mean Intersection-Over-Union metric for one-hot encoded or categorical labels.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        target_class_ids: A tuple or list of target class ids for which the metric is returned. 
    Returns:
        the fraction of correctly classified samples (float).
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    if y_true.ndim!=1:
        y_true = np.argmax(y_true, axis=-1)
    if y_pred.ndim!=1:
        y_pred = np.argmax(y_pred, axis=-1)
    sample_weight = _sample_weight(y_true, sample_weight)
    t = pd.DataFrame({'prob':y_pred.flatten(), 'label':y_true.flatten(), 'weight':sample_weight})
    if target_class_ids is None:
        target_class_ids = t.label.unique().tolist()
    result = {'IoU_mean':[], 'IoU_class':{i:0 for i in target_class_ids}}
    for i in t.label.unique():
        if i in target_class_ids:
            tp = (((t.label==i)&(t.prob==i))*t.weight).sum()
            fp = (((t.label==i)&(t.prob!=i))*t.weight).sum()
            fn = (((t.label!=i)&(t.prob==i))*t.weight).sum()
            result['IoU_mean'].append(tp/(tp+fp+fn))
            result['IoU_class'][i] = tp/(tp+fp+fn)
    result['IoU_mean'] = sum(result['IoU_mean'])/len(result['IoU_mean'])
    return result


def recall(y_true, y_pred, sample_weight=None, prob=0.5, pos_label=1):
    """Computes the recall of the predictions with respect to the labels.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        Recall of the positive class in binary classification.
    """
    sample_weight = _sample_weight(y_true, sample_weight)
    t = classified_func(y_true, y_pred, prob=prob, pos_label=pos_label)
    return (t.prob*sample_weight)[t.label==pos_label].mean()


def precision(y_true, y_pred, sample_weight=None, prob=0.5, pos_label=1):
    """Computes the precision of the predictions with respect to the labels.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        Precision of the positive class in binary classification.
    """
    sample_weight = _sample_weight(y_true, sample_weight)
    t = classified_func(y_true, y_pred, prob=prob, pos_label=pos_label)
    return (t.label*sample_weight)[t.prob==pos_label].mean()


def _cumsum_confusion_matrix(y_true, y_pred, sample_weight=None, pos_label=1):
    sample_weight = _sample_weight(y_true, sample_weight)
    t = pd.DataFrame({'prob':y_pred, 'label':y_true, 'weight':sample_weight})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    t = t.sort_values(['prob', 'label'], ascending=False).reset_index(drop=True)   
    t['tp'] = (t.label*t.weight).cumsum()
    t['fp'] = t.weight.cumsum()-t.tp
    t['tn'] = ((1-t.label)*t.weight).iloc[::-1].cumsum()
    t['fn'] = t.tp.max()-t.tp#t.weight.iloc[::-1].cumsum()-t.tn
    t['recall'] = t.tp/(t.tp + t.fn)
    t['precision'] = t.tp/(t.tp+t.fp)
    t['specificity'] = t.tn/(t.tn + t.fp)
    t['sensitivity'] = t.tp/(t.tp + t.fn)
    return t


def best_prob(y_true, y_pred, precision=None, recall=None, specificity=None, 
              accuracy=None, f1_score=None, iou_binary=None,
              sample_weight=None, pos_label=1):
    """Computes best best prob threshold when method is >= specified value.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        precision: A scalar value in range [0, 1].
        recall: A scalar value in range [0, 1].
        specificity: A scalar value in range [0, 1].
        accuracy: A scalar value in range [0, 1].
        f1_score: A scalar value in range [0, 1].
        iou_binary: A scalar value in range [0, 1].
        sample_weight: list or array or dict of sample weight.
        pos_label: positive label.
    Returns:
        best prob threshold.
    """
    t = _cumsum_confusion_matrix(y_true, y_pred, sample_weight=sample_weight, pos_label=pos_label)
    t['accuracy'] = (t.tp+t.tn)/(t.tp + t.fn+t.tn+t.fp)
    t['f1_score'] = 2*t.precision*t.recall/(t.precision+t.recall)
    t['iou_binary'] = t.tp/(t.tp+t.fp+t.fn)
    if recall is not None:
        t = t[t.recall>=recall]
    if precision is not None:
        t = t[t.precision>=precision]
    if specificity is not None:
        t = t[t.specificity>=specificity]
    if accuracy is not None:
        t = t[t.accuracy>=accuracy]
    if f1_score is not None:
        t = t[t.f1_score>=f1_score]
    if iou_binary is not None:
        t = t[t.iou_binary>=iou_binary]
    return t.prob.min()


def precision_on_recall(y_true, y_pred, recall=0.5, sample_weight=None, pos_label=1):
    """Computes best precision where recall is >= specified value.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        recall: A scalar value in range [0, 1].
        sample_weight: list or array or dict of sample weight.
        pos_label: positive label.
    Returns:
        prediction scores.
    """
    t = _cumsum_confusion_matrix(y_true, y_pred, sample_weight=sample_weight, pos_label=pos_label)
    return t[t.recall>=recall].precision.max()


def recall_on_precision(y_true, y_pred, precision=0.5, sample_weight=None, pos_label=1):
    """Computes best recall where precision is >= specified value.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        precision: A scalar value in range [0, 1].
        sample_weight: list or array or dict of sample weight.
        pos_label: positive label.
    Returns:
        recall scores.
    """
    t = _cumsum_confusion_matrix(y_true, y_pred, sample_weight=sample_weight, pos_label=pos_label)
    return t[t.precision>=precision].recall.max()


def sensitivity_on_specificity(y_true, y_pred, specificity=0.5, sample_weight=None, pos_label=1):
    """Computes best sensitivity(recall) where specificity is >= specified value.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        specificity: A scalar value in range [0, 1].
        sample_weight: list or array or dict of sample weight.
        pos_label: positive label.
    Returns:
        sensitivity scores.
    """
    t = _cumsum_confusion_matrix(y_true, y_pred, sample_weight=sample_weight, pos_label=pos_label)
    return t[t.specificity>=specificity].sensitivity.max()


def specificity_on_sensitivity(y_true, y_pred, sensitivity=0.5, sample_weight=None, pos_label=1):
    """Computes best specificity where sensitivity(recall) is >= specified value.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sensitivity: A scalar value in range [0, 1].
        sample_weight: list or array or dict of sample weight.
        pos_label: positive label.
    Returns:
        specificity scores.
    """
    t = _cumsum_confusion_matrix(y_true, y_pred, sample_weight=sample_weight, pos_label=pos_label)
    return t[t.sensitivity>=sensitivity].specificity.max()


def confusion_matrix(y_true, y_pred):
    """
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sample_weight: list or array of sample weight.
    Returns:
        Confusion matrix.
    """
    t = pd.DataFrame({'actual':y_true, 'predict':y_pred})
    t = pd.crosstab(t.predict, t.actual)
    return t


def fbeta_score(y_true, y_pred, beta, sample_weight=None, prob=0.5, pos_label=1):
    """
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        beta : int or float, weight of precision in harmonic mean.
        sample_weight: list or array or dict of sample weight.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        Fbeta score of the positive class in binary classification.
    """
    sample_weight = _sample_weight(y_true, sample_weight)
    r = recall(y_true, y_pred, sample_weight, prob, pos_label)
    p = precision(y_true, y_pred, sample_weight, prob, pos_label)
    return r*p*(1+np.power(beta, 2))/(np.power(beta, 2)*p+r)


def f1_score(y_true, y_pred, sample_weight=None, prob=0.5, pos_label=1):
    """
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        F1 score of the positive class in binary classification.
    """
    sample_weight = _sample_weight(y_true, sample_weight)
    return fbeta_score(y_true, y_pred, beta=1, sample_weight=sample_weight, prob=prob, pos_label=pos_label)


def auc_roc(y_true, y_pred, sample_weight=None, pos_label=1):
    """Area Under the Receiver Operating Characteristic Curve (ROC AUC)
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        pos_label: positive label.
    Returns:
        Area Under the Receiver Operating Characteristic Curve (ROC AUC) from prediction scores.
    """
#     sample_weight = _sample_weight(y_true, sample_weight)
#     t = pd.DataFrame({'prob':y_pred, 'label':y_true, 'weight':sample_weight})
#     assert t.label.nunique()==2, "`y_true` should be binary classification."
#     t.insert(0, 'target', t[t.label!=pos_label].label.unique()[0])
#     t = t[t.label!=pos_label].merge(t[t.label==pos_label], on='target')
#     auc = ((t.prob_y>t.prob_x)*(t.weight_y+t.weight_x)/2).mean()+((t.prob_y==t.prob_x)*(t.weight_y+t.weight_x)/2).mean()/2
    
    sample_weight = _sample_weight(y_true, sample_weight)
    t = pd.DataFrame({'prob':y_pred, 'label':y_true, 'weight':sample_weight}).sort_values(['prob']).reset_index(drop=True)
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    pos_rank = ((t[t.label==pos_label].index+1).values*(t[t.label==pos_label].weight.values)).sum()
    pos_cnt = t[t.label==pos_label].weight.sum()
    neg_cnt = t[t.label!=pos_label].weight.sum()
    auc = (pos_rank - pos_cnt*(pos_cnt+1)/2) / (pos_cnt*neg_cnt)
    return auc


def auc_pr(y_true, y_pred, sample_weight=None, pos_label=1):
    """Area Under the Receiver Operating Characteristic Curve (PR AUC)
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        pos_label: positive label.
    Returns:
        Area Under the Receiver Operating Characteristic Curve (PR AUC) from prediction scores.
    """
    sample_weight = _sample_weight(y_true, sample_weight)
    t = pd.DataFrame({'prob':y_pred, 'label':y_true, 'weight':sample_weight})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    t = t.sort_values(['prob', 'label'], ascending=False).reset_index(drop=True)   
    t['tp'] = (t.label*t.weight).cumsum()
    t['fp'] = t.weight.cumsum()-t.tp
    t['recall'] = t.tp/(t.label*t.weight).sum()
    t['precision'] = t.tp/(t.tp+t.fp)
    auc = t.sort_values(['recall', 'precision']).drop_duplicates(subset=['recall'], keep='last').precision.mean()
    return auc


def crossentropy_binary(y_true, y_pred, sample_weight=None):
    """Computes the crossentropy metric between the labels and predictions.
    
    This is the crossentropy metric class to be used when there are only two label classes (0 and 1).
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted probability, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
    Returns:
        binary crossentropy of the positive class in binary classification.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    t = np.exp((y_pred-np.max(y_pred))*sample_weight)
    t = -(np.log(t/t.sum())*y_true).mean()
    return t


def crossentropy_categorical(y_true, y_pred, sample_weight=None, one_hot=False):
    """Computes the crossentropy metric between the labels and predictions.
    
    This is the crossentropy metric class to be used when there are multiple label classes (2 or more). 
    Here we assume that labels are given as a one_hot representation. 
    eg., When labels values are [2, 0, 1], y_true = [[0, 0, 1], [1, 0, 0], [0, 1, 0]].
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted probability, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        one_hot: default True, Whether y_true is a one_hot variable.
    Returns:
        categorical crossentropy of the positive class in categorical classification.
    """
    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    sample_weight = _sample_weight(y_true, sample_weight)
    assert y_pred.shape[1]==np.unique(np.array(y_true)).size, "`y_pred` and `y_true` dim not same."
    t = np.exp(y_pred.T-np.max(y_pred, axis=1))
    if one_hot:
        t = -((np.log(t/np.sum(t, axis=0)).T*pd.get_dummies(y_true)).sum(axis=1)*sample_weight).mean()
    else:
        t = -((np.log(t/np.sum(t, axis=0)).T*y_true).sum(axis=1)*sample_weight).mean()
    return t


def ks(y_true, y_pred, sample_weight=None, pos_label=1):
    """Kolmogorov-Smirnov metrics.
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted probability, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        pos_label: positive label.
    Returns:
        KS score of the positive class in binary classification.
    """
    sample_weight = _sample_weight(y_true, sample_weight)
    t = pd.DataFrame({'prob':y_pred, 'label':y_true, 'weight':sample_weight})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    t = t.sort_values(['prob', 'label'], ascending=False).reset_index(drop=True)   
    t['tp'] = (t.label*t.weight).cumsum()
    t['fp'] = t.weight.cumsum()-t.tp
    t['tpr'] = t.tp/(t.label*t.weight).sum()
    t['fpr'] = t.fp/(t.weight.sum()-(t.label*t.weight).sum())
    ks = (t.tpr-t.fpr).abs().max()
    return ks


def gini(y_true, y_pred, sample_weight=None, pos_label=1):
    """Gini Coefficient
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted probability, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        pos_label: positive label.
    Returns:
        Gini score of the positive class in binary classification.
    """
    sample_weight = _sample_weight(y_true, sample_weight)
    t = pd.DataFrame({'prob':y_pred, 'label':y_true, 'weight':sample_weight})
    assert t.label.nunique()==2, "`y_true` should be binary classification."
    label_dict = {i:1 if i==pos_label else 0 for i in t.label.unique()}
    t['label'] = t.label.replace(label_dict)
    t = t.sort_values(['prob', 'label'], ascending=False).reset_index(drop=True)
    gini = ((t.label*t.weight).cumsum().sum()/(t.label*t.weight).sum()-(t.weight.sum()+1)/2)/t.weight.sum()
    return gini


def psi(y_true, y_pred, threshold):
    """population stability index
    
    Args:
        y_true: pd.Series or array or list, a feature variable.
        y_pred: pd.Series or array or list, a feature variable.
        threshold: list, a threshold list.
    Returns:
        psi value of two variable.
    """
    y_true = pd.Series(y_true)
    y_pred = pd.Series(y_pred)
    actual = (y_true-y_true.min())/(y_true.max()-y_true.min())
    predict = (y_pred-y_pred.min())/(y_pred.max()-y_pred.min())
    actual = pd.cut(actual, threshold, labels=range(1, len(threshold))).value_counts(normalize=True).reset_index()
    actual.columns = ['label', 'prob1']
    predict = pd.cut(predict, threshold, labels=range(1, len(threshold))).value_counts(normalize=True).reset_index()
    predict.columns = ['label', 'prob2']
    predict = actual.merge(predict, on='label', how='outer')
    psi = ((predict.prob1-predict.prob2)*np.log((predict.prob1/(predict.prob2+0.00000001)))).sum()
    return psi


def fmi(y_true, y_pred, sample_weight=None, prob=0.5, pos_label=1):
    """
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        prob: probability threshold.
        pos_label: positive label.
    Returns:
        FMI of the positive class in binary classification.
    """
    sample_weight = _sample_weight(y_true, sample_weight)
    t = classified_func(y_true, y_pred, prob=prob, pos_label=pos_label)
    t['weight'] = sample_weight
    tp = (((t.label==pos_label)&(t.prob==pos_label))*t.weight).sum()
    fp = (((t.label==pos_label)&(t.prob!=pos_label))*t.weight).sum()
    fn = (((t.label!=pos_label)&(t.prob==pos_label))*t.weight).sum()
    return tp/np.sqrt((tp+fp)*(tp+fn))


def report_binary(y_true, y_pred, sample_weight=None, prob=0.5, pos_label=1, printable=False):
    """binary metrics report
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted labels, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        prob: probability threshold.
        pos_label: positive label.
        printable: bool, print report.
    Returns:
        binary report of the positive class in binary classification.
    """
    t = classified_func(y_true, y_pred, prob=prob, pos_label=pos_label)
    result = {'accuracy':round(accuracy_binary(y_true, y_pred, sample_weight=sample_weight, prob=prob, pos_label=pos_label), 4),
              'precision':round(precision(y_true, y_pred, sample_weight=sample_weight, prob=prob, pos_label=pos_label), 4),
              'recall':round(recall(y_true, y_pred, sample_weight=sample_weight, prob=prob, pos_label=pos_label), 4),
              'f1_score':round(f1_score(y_true, y_pred, sample_weight=sample_weight, prob=prob, pos_label=pos_label), 4),
              'auc_roc':round(auc_roc(y_true, y_pred, sample_weight=sample_weight, pos_label=pos_label), 4),
              'auc_pr':round(auc_pr(y_true, y_pred, sample_weight=sample_weight, pos_label=pos_label), 4),
              'fmi':round(fmi(y_true, y_pred, sample_weight=sample_weight, prob=prob, pos_label=pos_label), 4),
              'gini':round(gini(y_true, y_pred, sample_weight=sample_weight, pos_label=pos_label), 4),
              'ks':round(ks(y_true, y_pred, sample_weight=sample_weight, pos_label=pos_label), 4),
              'RMSE_label':round(mean_squared_error(t.label, t.prob, log=False, root=True, sample_weight=sample_weight), 4),
              'RMSE_prob':round(mean_squared_error(y_true, y_pred, log=False, root=True, sample_weight=sample_weight), 4)
             }
    if printable:
        print("\nBinary Classification Report")
        print("Accuracy: %.4f" % result['accuracy'])
        print("Precision: %.4f" % result['precision'])
        print("Recall: %.4f" % result['recall'])
        print("F1_score: %.4f" % result['f1_score'])
        print("AUC_ROC Score: %.4f" % result['auc_roc'])
        print("AUC_PR Score: %.4f" % result['auc_pr'])
        print("FMI: %.4f" % result['fmi'])
        print("KS: %.4f" % result['ks'])
        print("Gini: %.4f" % result['gini'])
        print("RMSE_label: %.4f" % result['RMSE_label'])
        print("RMSE_prob: %.4f" % result['RMSE_prob'])
    return result


def accuracy_categorical_top_k(y_true, y_pred, k, sample_weight=None):
    """top k categorical accuracy
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted probs, as returned by a classifier.
        sample_weight: list or array or dict of sample weight.
        k: Number of top elements to look at for computing accuracy.
    Returns:
        Top K categorical accuracy value.
    """
    y_true = np.array(y_true)
    if y_true.ndim!=1:
        y_true = np.argmax(y_true, axis=-1)
    sample_weight = _sample_weight(y_true, sample_weight)
    y_pred = [[s.index(i) for i in sorted(s)[-k:]] for s in y_pred]
    return np.mean([sample_weight[i] if j in y_pred[i] else 0 for i,j in enumerate(y_true)])



import pandas as pd

from linora.metrics._utils import _sample_weight

__all__  = ['mapk', 'hit_ratio', 'mean_reciprocal_rank']


def mapk(y_true, y_pred, k, sample_weight=None):
    """Mean Average Precision k
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a rank.
        k: int, top k predict values.
        sample_weight: list or array of sample weight.
    Returns:
        Mean Average Precision k values.
    """
    def apk(actual, predict, weight, k):
        if len(predict)>k:
            predict = predict[:k]
        score = 0.0
        nums = 0.0
        for i,p in enumerate(predict):
            if p in actual and p not in predict[:i]:
                nums += 1.0
                score += nums / (i+1.0)
        return score / min(len(actual), k)*weight if actual else 0.0
    sample_weight = _sample_weight(y_true, sample_weight)
    return pd.DataFrame({'label1':y_true, 'label2':y_pred, 'weight':sample_weight}).apply(lambda x:apk(x[0], x[1], x[2], k=k), axis=1).mean()


def hit_ratio(y_true, y_pred, k, sample_weight=None):
    """Hit Ratio k
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a rank.
        k: int, top k predict values.
        sample_weight: list or array of sample weight.
    Returns:
        Hit Ratio k values.
    """
    sample_weight = _sample_weight(y_true, sample_weight)
    t = pd.DataFrame({'label1':y_true, 'label2':y_pred, 'weight':sample_weight})
    return t.apply(lambda x:len(set(x[0]).intersection(set(x[1][:k])))*x[2], axis=1).sum()/t.label1.map(lambda x:len(set(x))).sum()


def mean_reciprocal_rank(y_true, y_pred, k, sample_weight=None):
    """Mean Reciprocal Rank
    
    Args:
        y_true: pd.Series or array or list, ground truth (correct) labels.
        y_pred: pd.Series or array or list, predicted values, as returned by a rank.
        k: int, top k predict values.
        sample_weight: list or array of sample weight.
    Returns:
        mean reciprocal rank k values.
    """
    def mrr(actual, predict, weight, k):
        try:
            rank = 1./(predict[:k].index(actual)+1)*weight
        except:
            rank = 0
        return rank
    sample_weight = _sample_weight(y_true, sample_weight)
    return pd.DataFrame({'label1':y_true, 'label2':y_pred, 'weight':sample_weight}).apply(lambda x: mrr(x[0], x[1], x[2], k=k), axis=1).mean()



from linora.feature_column._categorical import *

class FeatureCategorical(object):
    def __init__(self):
        self.pipe = {}
        
    def categorical_count(self, variable, normalize=True, abnormal_value=0, miss_value=0, name=None, keep=True):
        """Count or frequency of conversion category variables.
    
        Args:
            variable: str, feature variable name.
            normalize: bool, If True then the object returned will contain the relative frequencies of the unique values.
            abnormal_value: int or float, if feature values not in feature_scale dict, return `abnormal_value`.
            miss_value: int or float, if feature values are missing, return `miss_value`.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'normalize':normalize, 'abnormal_value':abnormal_value, 
                           'miss_value':miss_value, 'name':variable if name is None else name},
                  'type':'categorical_count', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def categorical_crossed(self, variable, hash_bucket_size=3, name=None, keep=True):
        """Crossed categories and hash labels with value between 0 and hash_bucket_size-1.

        Args:
            variable: list, feature variable name of list.
            hash_bucket_size: int, number of categories that need hash.
            name: str, output feature name, if None, name is feature.name .
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'hash_bucket_size':hash_bucket_size, 
                           'name':'_'.join(name)+'_crossed' if name is None else name}, 
                  'type':'categorical_crossed', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def categorical_encoder(self, variable, abnormal_value=-1, miss_value=-1, name=None, keep=True):
        """Encode labels with value between 0 and n_classes-1.

        Args:
            variable: str, feature variable name.
            abnormal_value: int, if feature values not in feature_scale dict, return `abnormal_value`.
            miss_value: int or float, if feature values are missing, return `miss_value`.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'abnormal_value':abnormal_value, 'miss_value':miss_value, 
                           'name':variable if name is None else name},
                  'type':'categorical_encoder', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def categorical_hash(self, variable, hash_bucket_size=3, name=None, keep=True):
        """Hash labels with value between 0 and hash_bucket_size-1.

        Args:
            variable: str, feature variable name.
            hash_bucket_size: int, number of categories that need hash.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'hash_bucket_size':hash_bucket_size, 
                           'name':variable if name is None else name},
                  'type':'categorical_hash', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def categorical_hist(self, variable, label, abnormal_value=0, miss_value=0, name=None, keep=True):
        """Hist labels with value counts prob.

        Args:
            feature: str, feature variable name.
            label: str, label variable name.
            abnormal_value: int, if feature values not in feature_scale dict, return `abnormal_value`.
            miss_value: int or float, if feature values are missing, return `miss_value`.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'abnormal_value':abnormal_value, 'miss_value':miss_value, 
                           'name':variable if name is None else name, 'label':label},
                  'type':'categorical_hist', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def categorical_onehot_binarizer(self, variable, abnormal_value=0, miss_value=0, name=None, keep=True):
        """Transform between iterable of iterables and a multilabel format, sample is simple categories.

        Args:
            variable: str, feature variable name.
            abnormal_value: int, if feature values not in feature_scale dict, return `abnormal_value`.
            miss_value: int or float, if feature values are missing, return `miss_value`.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'abnormal_value':abnormal_value, 'miss_value':miss_value, 
                           'name':variable if name is None else name},
                  'type':'categorical_onehot_binarizer', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def categorical_onehot_multiple(self, variable, abnormal_value=0, miss_value=0, name=None, keep=True):
        """Transform between iterable of iterables and a multilabel format, sample is multiple categories.

        Args:
            variable: str, feature variable name.
            abnormal_value: int, if feature values not in feature_scale dict, return `abnormal_value`.
            miss_value: int or float, if feature values are missing, return `miss_value`.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'abnormal_value':abnormal_value, 'miss_value':miss_value, 
                           'name':variable if name is None else name},
                  'type':'categorical_onehot_multiple', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def categorical_rare(self, variable, p=0.05, min_num=None, max_num=None, abnormal_value=-1, miss_value=-1, name=None, keep=True):
        """Groups rare or infrequent categories in a new category called “Rare”, or any other name entered by the user.

        Args:
            variable: str, feature variable name.
            p: The minimum frequency a label should have to be considered frequent. 
                Categories with frequencies lower than tol will be grouped.
            min_num: The minimum number of categories a variable should have for the encoder to find frequent labels. 
                If the variable contains less categories, all of them will be considered frequent.
            max_num: The maximum number of categories that should be considered frequent. 
                If None, all categories with frequency above the tolerance (tol) will be considered frequent. 
                If you enter 5, only the 4 most frequent categories will be retained and the rest grouped.
            abnormal_value: int or float, if feature values not in feature_scale dict, return `abnormal_value`.
            miss_value: int or float, if feature values are missing, return `miss_value`.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'p':p, 'min_num':min_num, 'max_num':max_num,
                           'abnormal_value':abnormal_value, 'miss_value':miss_value, 
                           'name':variable if name is None else name},
                  'type':'categorical_rare', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def categorical_regress(self, variable, label, method='mean', abnormal_value='mean', miss_value='mean', name=None, keep=True):
        """Regress labels with value counts prob.

        Args:
            variable: str, feature variable name.
            label: str, label variable name.
            method: 'mean' or 'median'
            abnormal_value: int, if feature values not in feature_scale dict, return `abnormal_value`.
            miss_value: int or float, if feature values are missing, return `miss_value`.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'method':method,
                           'abnormal_value':abnormal_value, 'miss_value':miss_value, 
                           'name':variable if name is None else name, 'label':label},
                  'type':'categorical_regress', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def categorical_woe(self, variable, label, pos_label=1, abnormal_value=-1, miss_value=-1, name=None, keep=True):
        """Calculate series woe value

        Args:
            variable: str, feature variable name.
            label: str, label variable name.
            pos_label: int, default=1, positive label value.
            abnormal_value: int, if feature values not in feature_scale dict, return `abnormal_value`.
            miss_value: int or float, if feature values are missing, return `miss_value`.
            name: str, output feature name, if None, name is variable.
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'pos_label':pos_label,
                           'abnormal_value':abnormal_value, 'miss_value':miss_value, 
                           'name':variable if name is None else name, 'label':label},
                  'type':'categorical_woe', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    
    


from PIL import Image

__all__ = ['ResizeMode', 'resize']


class resize_method:
    """For geometry operations that may map multiple input pixels to a single output pixel, 
    the Python Imaging Library provides different resampling filters.
    
    NEAREST: Pick one nearest pixel from the input image. Ignore all other input pixels.
    BOX: Each pixel of source image contributes to one pixel of the destination 
         image with identical weights. 
    BILINEAR: For resize calculate the output pixel value using linear interpolation on all 
              pixels that may contribute to the output value. 
              For other transformations linear interpolation over a 2x2 environment 
              in the input image is used.
    HAMMING: Produces a sharper image than BILINEAR, 
             doesn’t have dislocations on local level like with BOX.
    BICUBIC: For resize calculate the output pixel value using cubic interpolation on all 
             pixels that may contribute to the output value. For other transformations cubic 
             interpolation over a 4x4 environment in the input image is used.
    LANCZOS: Calculate the output pixel value using a high-quality Lanczos filter (a truncated sinc) 
             on all pixels that may contribute to the output value.
    """
    NEAREST = Image.Resampling.NEAREST
    BOX = Image.Resampling.BOX
    BILINEAR = Image.Resampling.BILINEAR
    HAMMING = Image.Resampling.HAMMING
    BICUBIC = Image.Resampling.BICUBIC
    LANCZOS = Image.Resampling.LANCZOS
    
ResizeMode = resize_method()


def resize(image, size, method=ResizeMode.BILINEAR):
    """Returns a resized copy of this image.
    
    Args:
        image: a PIL instance.
        size: The requested size in pixels, as a 2-tuple: (width, height).
              if float, (0,1], size=[int(width*size), int(height*size)]
              if [0.2, 0.7], size=[int(width*0.2), int(height*0.7)]
        method: An optional resampling filter. see la.image.ResizeMode.
    Returns:
        a PIL instance.
    """
    if image.mode in ['1', 'P'] and method!=ResizeMode.NEAREST:
        method = ResizeMode.NEAREST
    if isinstance(size, float):
        size = (int(image.size[0]*size), int(image.size[1]*size))
    elif isinstance(size, (list, tuple)):
        if isinstance(size[0], float):
            size = (int(image.size[0]*size[0]), size[1])
        if isinstance(size[1], float):
            size = (size[0], int(image.size[1]*size[1]))
    return image.resize(size, resample=method)



import os
import json
from collections import Counter

import numpy as np
import pandas as pd

from linora.utils._config import Config
from linora.utils._logger import Logger
from linora.sample._fold import kfold, train_test_split
from linora.param_search._config import model_hp
from linora.param_search._config import __xgboost_version__, __lightgbm_version__


__all__ = ['RandomSearch', 'GridSearch']


class BaseSearch():
    def __init__(self, model, hp=None, name=None, method='random'):
        self.params = Config()
        if name is not None:
            self.params.name = name
        elif method=='random':
            self.params.name = 'RS'
        elif method=='grid':
            self.params.name = 'GS'
        self.params.model_init = model
        self.params.model_name = ''
        self.params.method = method
        if model in ['XGBClassifier', 'XGBRegressor', 'LGBMClassifier', 'LGBMRegressor']:
            self.hp = model_hp(model=model, method=method)
            if hp is not None:
                self.hp.from_HyperParameters(hp)
            self.params.model_name = model
            if model=='XGBClassifier':
                import xgboost as xgb
                assert xgb.__version__>=__xgboost_version__, f'xgboost version should be >={__xgboost_version__}.'
                self.params.model_init = xgb.XGBClassifier
            elif model=='LGBMClassifier':
                import xgboost as xgb
                assert xgb.__version__>=__xgboost_version__, f'xgboost version should be >={__xgboost_version__}.'
                self.params.model_init = xgb.XGBRegressor
            elif model=='LGBMClassifier':
                import lightgbm as lgb
                assert lgb.__version__>=__lightgbm_version__, f'lightgbm version should be >={__lightgbm_version__}.'
                self.params.model_init = lgb.LGBMClassifier
            elif model=='LGBMRegressor':
                import lightgbm as lgb
                assert lgb.__version__>=__lightgbm_version__, f'lightgbm version should be >={__lightgbm_version__}.'
                self.params.model_init = lgb.LGBMRegressor
        else:
            self.hp = hp
        self.best_params = dict()
        self.best_params_history = dict()

    def search(self, train_data, metrics, valid_data=None,
               iter_num=None, cv=3, metrics_min=True, 
               speedy=True, speedy_param=(20000, 0.3), 
               save_model_dir=None, save_model_name=None):
        """model params search method.

        Args:
            train_data: A list of (X, y, sample_weight) tuple pairs to use as train sets.
            metrics: model metrics function.
            valid_data: A list of (X, y, sample_weight) tuple pairs to use as validation sets.
            iter_num: search count.
            cv: cross validation fold.
            metrics_min: metrics value whether the smaller the better.
            speedy: whether use speedy method.
            speedy_param: if use speedy method, test_size will be set, 
                          test_size = 1-round(min(speedy_param[0], feature.shape[0]*speedy_param[1])/feature.shape[0], 2).
            save_model_dir: str, save model folder, only work with model='XGBClassifier' or 'XGBRegressor'.
            save_model_name: str, save model name prefix, only work with model='XGBClassifier' or 'XGBRegressor'.
        Returns:
            a best model params dict.
        Raises:
            params error.
        """
        logger = Logger(name=self.params.name)
        logger.info(f"Start hyperparameter {self.params.method} search.")
        import warnings
        warnings.filterwarnings("ignore")
        if speedy:
            test_size = 1-round(min(speedy_param[0], len(train_data[1])*speedy_param[1])/len(train_data[1]), 2)
        if self.params.model_name=='XGBClassifier':
            self._xgb_weight(train_data[1])
        
        if valid_data is not None:
            cv_score_list = []
            
        if self.params.method=='grid':
            if iter_num is None:
                iter_num = self.hp.cardinality()
            else:
                iter_num = min(iter_num, self.hp.cardinality())
        if iter_num is None:
            iter_num = 100
        for i in range(1, iter_num+1):
            self.hp.update(self.best_params)
            self.params.model = self.params.model_init(**self.hp.params)
            score = []
            if speedy:
                for _ in range(cv):
                    index = train_test_split(train_data[0], train_data[1], test_size, seed=np.random.choice(range(100), 1)[0])
                    score.append(self._model_fit_predict(train_data, metrics, index, mode=1))
            else:
                index_list = kfold(train_data[0], train_data[1], n_splits=cv, seed=np.random.choice(range(100), 1)[0])
                for n, index in enumerate(index_list):
                    score.append(self._model_fit_predict(train_data, metrics, index, mode=1))
            cv_score = np.mean(score)
            if valid_data is not None:
                cv_score_list.append(cv_score)
                cv_score_list.sort()
                threshold = cv_score_list[int(len(cv_score_list)*(0.2 if metrics_min else 0.8))]
                if (metrics_min==True and threshold>=cv_score) or (metrics_min==False and threshold<=cv_score):
                    cv_score = self._model_fit_predict(valid_data, metrics, index=None, mode=0)
                else:
                    logger.info(f"Model {self.params.method} search progress: {i/iter_num*100:.1f}%, best score: {scoring:.4f}", enter=False if i<iter_num else True)
                    continue
            if i==1:
                scoring = cv_score
            if (metrics_min==True and cv_score<=scoring) or (metrics_min==False and cv_score>=scoring):
                scoring = cv_score
                self.best_params = self.hp.params.copy()
                self.best_params_history[i] = {'score':scoring, 'best_params':self.best_params.copy()}
                if self.params.model_name in ['XGBClassifier', 'XGBRegressor']:
                    if save_model_dir is not None:
                        if save_model_name is None:
                            save_model_name = self.params.name
                        model.save_model(os.path.join(save_model_dir, f"{save_model_name}_model.json"))
                        with open(os.path.join(save_model_dir, f"{save_model_name}_params.json"),'w') as f:
                            json.dump(best_params, f)
            logger.info(f"Model {self.params.method} search progress: {i/iter_num*100:.1f}%, best score: {scoring:.4f}", enter=False if i<iter_num else True)
        logger.info(f"Model {self.params.method} search best score: {scoring:.4f}", close=True, time_mode=1)
        return self.best_params

    def _model_fit_predict(self, data, metrics, index=None, mode=1):
        if mode:
            if len(data)==2:
                self.params.model.fit(data[0].loc[index[0]], data[1][index[0]])
            else:
                self.params.model.fit(data[0].loc[index[0]], data[1][index[0]], sample_weight=data[2][index[0]])
        if index is None:
            cv_pred = pd.Series(self.params.model.predict(data[0]), index=data[1].index)
        else:
            cv_pred = pd.Series(self.params.model.predict(data[0].loc[index[1]]), index=data[1][index[1]].index)
        if len(data)==2:
            if index is None:
                return metrics(data[1], cv_pred)
            else:
                return metrics(data[1][index[1]], cv_pred)
        else:
            if index is None:
                return metrics(data[1], cv_pred, sample_weight=data[2])
            else:
                return metrics(data[1][index[1]], cv_pred, sample_weight=data[2][index[1]])
    
    def _xgb_weight(self, label):
        weight_dict = Counter(label)
        if len(weight_dict)==2:
            weight = int(np.ceil(weight_dict[min(weight_dict)]/weight_dict[max(weight_dict)]))
        else:
            weight_dict = {j:i for i,j in weight_dict.items()}
            weight = int(np.ceil(weight_dict[max(weight_dict)]/weight_dict[min(weight_dict)]))
        if self.params.method=='grid':
            self.hp.Choice('scale_pos_weight', [1, weight], weight, rank=6)
        else:
            self.hp.Choice('scale_pos_weight', [1, weight])
        
        
class RandomSearch(BaseSearch):
    """model params search use RandomSearch method.
    
    Args:
        model: a model object or one of ['XGBClassifier', 'XGBRegressor', 'LGBMClassifier', 'LGBMRegressor'].
        hp: a la.param_search.HyperParametersRandom object.
        name: logger name.
    """
    def __init__(self, model, hp=None, name=None):
        super(RandomSearch, self).__init__(model, hp=hp, name=name, method='random')


class GridSearch(BaseSearch):
    """model params search use GridSearch method.
    
    Args:
        model: a model object or one of ['XGBClassifier', 'XGBRegressor', 'LGBMClassifier', 'LGBMRegressor'].
        hp: a la.param_search.HyperParametersGrid object.
        name: logger name.
    """
    def __init__(self, model, hp=None, name=None):
        super(GridSearch, self).__init__(model, hp=hp, name=name, method='grid')


from linora.feature_column._categorical import *
from linora.feature_column._datetime import *
from linora.feature_column._feature import *
from linora.feature_column._normalize import *
from linora.feature_column._numerical import *
from linora.feature_column import boundary



import os
import pathlib
import random as random1
from itertools import chain

import numpy as np
import pandas as pd

from linora import gfile
from linora.data.Dataset._data import DataSet
from linora.parallel._thread import ThreadLoom

__all__ = ['from_tensor', 'from_folder', 'from_class_folder', 'range', 'random', 
           'choose_from_datasets', 'sample_from_datasets']


class BatchFunction():
    def _batch_list_map(self, loc):
        if 'array' in self._params.data_mode:
            data = list(map(self._params.map[self._params.mode][0], *(i[loc] for i in self._params.data[self._params.mode1])))
            return [np.array(list(map(lambda x:x[i], data))) for i in np.arange(len(data[0]))]
        loom = ThreadLoom(self._params.map[self._params.mode][1])
        for i in loc:
            loom.add_function(self._params.map[self._params.mode][0], [j[i] for j in self._params.data[self._params.mode1]])
        t = loom.execute()
        for i in t:
            if t[i]['got_error']:
                continue
            if isinstance(t[i]['output'], (list, tuple)):
                return [np.concatenate([np.expand_dims(t[j]['output'][k], 0) for j in t if not t[j]['got_error']]) for k in np.arange(len(t[i]['output']))]
            else:
                return np.concatenate([np.expand_dims(t[j]['output'], 0) for j in t if not t[j]['got_error']])
    
    def _batch_list(self, loc):
        return [i[loc] for i in self._params.data[self._params.mode1]]
    
    def _batch_map(self, loc):
        if 'array' in self._params.data_mode:
            return np.array(list(map(self._params.map[self._params.mode][0], self._params.data[self._params.mode1][loc])))
        loom = ThreadLoom(self._params.map[self._params.mode][1])
        for i in loc:
            loom.add_function(self._params.map[self._params.mode][0], [self._params.data[self._params.mode1][i]])
        t = loom.execute()
        for i in t:
            if t[i]['got_error']:
                continue
            if isinstance(t[i]['output'], (list, tuple)):
                return [np.concatenate([np.expand_dims(t[j]['output'][k], 0) for j in t if not t[j]['got_error']]) for k in np.arange(len(t[i]['output']))]
            else:
                return np.concatenate([np.expand_dims(t[j]['output'], 0) for j in t if not t[j]['got_error']])
    
    def _batch(self, loc):
        return self._params.data[self._params.mode1][loc]


class from_tensor(DataSet, BatchFunction):
    """Creates a Dataset whose elements are slices of the given tensors.
        
        Args:
            data: numpy array or tuple of numpy array
    """
    def __init__(self, data):
        super(from_tensor, self).__init__()
        if isinstance(data, (int, float, str)):
            self._params.data[self._params.mode1] = np.array([data])
        elif isinstance(data, tuple):
            for i in data:
                assert len(data[0])==len(i), 'Length needs to be consistent between data.'
            self._params.data[self._params.mode1] = [np.array(i) for i in data]
        else:
            self._params.data[self._params.mode1] = np.array(data)
        self._params.index[self._params.mode] = list(np.arange(len(self._params.data[self._params.mode1][0] if isinstance(data, tuple) else self._params.data[self._params.mode1])))
        self._data_mode()
    

class from_folder(DataSet, BatchFunction):
    """Construct an image dataset label index.

    Args:
        root: image dataset file root.
        image_format: list, default ['png', 'jpg', 'jpeg'].
        label_func: Function is applied to the name of each picture.
        label_encoder: whether encode labels with value between 0 and n_classes-1.
    """
    def __init__(self, root, image_format=['png', 'jpg', 'jpeg'], label_func=None, label_encoder=False):
        super(from_folder, self).__init__()
        p = pathlib.Path(root)
        dataset = pd.DataFrame(chain.from_iterable((p.rglob(f'*.{i}') for i in image_format)), columns=['image'])
        if label_func is not None:
            dataset['label'] = dataset.image.map(lambda x:label_func(x.name))
            name_label_dict = {j: i for i, j in enumerate(dataset.label.unique())}
            self.name_label_dict = {'positive':name_label_dict, 'negative':{j: i for i, j in name_label_dict.items()}}
            if label_encoder:
                dataset['label'] = dataset.label.replace(self.name_label_dict['positive'])
        dataset['image'] = dataset.image.astype(str).map(lambda x:eval(repr(x).replace("\\", '/').replace("//", '/')))
        if label_func is not None:
            self._params.data[self._params.mode1] = [dataset.image.values, dataset.label.values]
        else:
            self._params.data[self._params.mode1] = dataset.image.values
        self._params.index[self._params.mode] = dataset.index.to_list()
        self._data_mode()
        self._params.data_from = 'from_folder'
        
    
class from_class_folder(DataSet, BatchFunction):
    """Construct an image dataset label index.

    Args:
        root: image dataset file root.
        image_format: list, default ['png', 'jpg', 'jpeg'].
        label_encoder: whether encode labels with value between 0 and n_classes-1.
    """
    def __init__(self, root, image_format=['png', 'jpg', 'jpeg'], label_encoder=False):
        super(from_class_folder, self).__init__()
        file = os.listdir(root)
        file = [i for i in file if os.path.isdir(root+'/'+i) and i[0]!='.']
        data = pd.DataFrame()
        for i in file:
            data = pd.concat([data, pd.DataFrame({'image':os.listdir(root+'/'+i), 'label':i})])
        data = data.reset_index(drop=True)
        data['image'] = root+'/'+data.label+'/'+data.image
        data = data[data.image.map(lambda x: True if '.' in x.split('/')[-1] else False)]
        data = data[data.image.map(lambda x: True if x.split('/')[-1][0]!='.' else False)]
        data = data[data.image.map(lambda x: True if len(x.split('/')[-1].split('.'))==2 else False)]
        data = data[data.image.map(lambda x: True if str.lower(x.split('/')[-1].split('.')[1]) in image_format else False)]
        data = data.reset_index(drop=True)
        name_label_dict = {j: i for i, j in enumerate(data.label.unique())}
        self.name_label_dict = {'positive':name_label_dict, 'negative':{j: i for i, j in name_label_dict.items()}}
        if label_encoder:
            data['label'] = data.label.replace(self.name_label_dict['positive'])
        self._params.data[self._params.mode1] = [data.image.values, data.label.values]
        self._params.index[self._params.mode] = data.index.to_list()
        self._data_mode()
        self._params.data_from = 'from_class_folder'


class range(DataSet, BatchFunction):
    """Creates a Dataset of a step-separated range of values."""
    def __init__(self, *args, **kwargs):
        super(range, self).__init__()
        self._params.data[self._params.mode1] = np.arange(*args, **kwargs)
        self._params.index[self._params.mode] = list(np.arange(len(self._params.data[self._params.mode1])))
        self._data_mode()


class random(DataSet, BatchFunction):
    """Creates a Dataset of pseudorandom values. The dataset generates a sequence of uniformly distributed integer values.
        
        Args:
            size: shape of output values.
            lower: min random values.
            upper: max random values.
            seed: random seed.
        """
    def __init__(self, size, lower=0, upper=10, seed=None):
        super(random, self).__init__()
        if isinstance(size, int):
            t = (list(np.arange(lower, upper))*(size//int(upper-lower)+1))[:size]
        else:
            t = 1
            for i in size:
                t *= i
            t = (list(np.arange(lower, upper))*(t//int(upper-lower)+1))[:t]
        random1.shuffle(t, random=lambda :((seed if seed is not None else random1.randint(1, 99)))%10/10)
        self._params.data[self._params.mode1] = np.array(t).reshape(size)
        self._params.index[self._params.mode] = list(np.arange(len(self._params.data[self._params.mode1])))
        self._data_mode()

        
class choose_from_datasets(DataSet, BatchFunction):
    """Creates a dataset that deterministically chooses elements from datasets.

    Args:
        datasets: A non-empty list of la.data.Dataset objects with compatible structure.
        index: A list of scalar between 0 and len(datasets) - 1.
        stop_on_empty_dataset: If True, selection stops if it encounters an empty dataset. 
                               If False, it skips empty datasets. It is recommended to set it to True. 
                               Otherwise, the selected elements start off as the user intends, 
                               but may change as input datasets become empty. 
                               This can be difficult to detect since the dataset starts off looking correct. 
                               Defaults to True.
    """
    def __init__(self, datasets, index, stop_on_empty_dataset=True):
        super(choose_from_datasets, self).__init__()
        if isinstance(datasets[0]._params.data[datasets[0]._params.mode1], list):
            self._params.data[self._params.mode1] = []
            for i in np.arange(len(datasets[0]._params.data[datasets[0]._params.mode1])):
                self._params.data[self._params.mode1] += [np.concatenate([sets._params.data[sets._params.mode1][i] for sets in datasets])]
        else:
            self._params.data[self._params.mode1] = np.concatenate([sets._params.data[sets._params.mode1] for sets in datasets])
        data_dict = []
        for r, sets in enumerate(datasets):
            if r==0:
                data_dict.append(sets._params.index[sets._params.mode].copy())
            else:
                t = max(data_dict[-1])+1
                data_dict.append([i+t for i in sets._params.index[sets._params.mode].copy()])
        if stop_on_empty_dataset:
            self._params.index[self._params.mode] = []
            for i in index:
                if len(data_dict[i])==0:
                    break
                self._params.index[self._params.mode].append(data_dict[i].pop(0))
        else:
            self._params.index[self._params.mode] = [data_dict[i].pop(0) for i in index if len(data_dict[i])>0]
        self._data_mode()
        
class sample_from_datasets(DataSet, BatchFunction):
    """Creates a dataset that not deterministically chooses elements from datasets.

    Args:
        datasets: A non-empty list of la.data.Dataset objects with compatible structure.
        weight: A list of len(datasets) floating-point values where weight[i] 
                represents the probability to sample from datasets[i].
        stop_on_empty_dataset: If True, selection stops if it encounters an empty dataset. 
                               If False, it skips empty datasets. It is recommended to set it to True. 
                               Otherwise, the selected elements start off as the user intends, 
                               but may change as input datasets become empty. 
                               This can be difficult to detect since the dataset starts off looking correct. 
    """
    def __init__(self, datasets, weight=None, stop_on_empty_dataset=False):
        super(sample_from_datasets, self).__init__()
        weights = weight.copy()
        if isinstance(datasets[0]._params.data[datasets[0]._params.mode1], list):
            self._params.data[self._params.mode1] = []
            for i in np.arange(len(datasets[0]._params.data[datasets[0]._params.mode1])):
                self._params.data[self._params.mode1] += [np.concatenate([sets._params.data[sets._params.mode1][i] for sets in datasets])]
        else:
            self._params.data[self._params.mode1] = np.concatenate([sets._params.data[sets._params.mode1] for sets in datasets])
        data_dict = []
        for r, sets in enumerate(datasets):
            if r==0:
                data_dict.append(sets._params.index[sets._params.mode].copy())
            else:
                t = max(data_dict[-1])+1
                data_dict.append([i+t for i in sets._params.index[sets._params.mode].copy()])
        if stop_on_empty_dataset:
            self._params.index[self._params.mode] = []
            while 1:
                index = np.random.choice(np.arange(len(data_dict)), p=weights)
                self._params.index[self._params.mode].append(data_dict[index].pop(0))
                if len(data_dict[index])==0:
                    break
        else:
            self._params.index[self._params.mode] = []
            while 1:
                if len(data_dict)==0:
                    break
                index = np.random.choice(np.arange(len(data_dict)), p=weights)
                self._params.index[self._params.mode].append(data_dict[index].pop(0))
                if len(data_dict[index])==0:
                    data_dict.pop(index)
                    if weights is not None:
                        weights.pop(index)
                        weights = [i/sum(weights) for i in weights]
        self._data_mode()


import datetime
import multiprocessing
from collections import defaultdict

from linora.utils._config import Config
from linora.utils._logger import Logger

__all__ = ['Schedulers']


class Schedulers():
    """Time job task manager."""
    def __init__(self, logger=None, verbose=0, config_file=None):
        """
        Args:
            logger: Logger object, linora.utils.Logger() class.
            verbose: Verbosity mode, 0 (silent), 1 (verbose).
            config_file: job task config file, if .py file.
                         example: .py file name is schedulers_config.py, contain a dict, 
                         config = {'hhh':{'mode':'every_minute', 'time':50, 'function':function, 'args':[], 'kwargs':{}}}
        """
        warn = Logger(name='')
        warn.info(f"la.utils.Schedulers has been deprecated and will be deleted in version 2.1.0. Please use la.server.Schedulers")
        self.config = dict()
        self.params = Config()
        self.params.verbose = verbose
        if logger is None:
            logger = Logger()
        self.params.logger = logger
        self.params.config_file = config_file
        manager = multiprocessing.Manager()
        self.params.tracker_dict = manager.dict()
        self.params.runner_dict = defaultdict()
    
    def every_minute(self, time, function, args=None, kwargs=None, name=None):
        """Run task manager every minute.
        
        Args:
            time: int or str, Within range 0~59, 27 means the task will start at the 27th second every minute. 
            function: task function.
            args: list, function args.
            kwargs: dict, function kwargs.
            name: task name, if None, is function name.
        """
        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()
        if name is None:
            name = function.__name__+(f'_{len(self.config)+1}' if function.__name__ in self.config else '')
        self.config[name] = {'mode':'every_minute', 'time':int(time), 'function':function, 'args':args, 
                             'kwargs':kwargs, 'execute_num':0, 'runner':(function, args, kwargs, name),
                             'time_init':datetime.datetime.now()}
        self.params.tracker_dict[name] = dict()
    
    def every_hour(self, time, function, args=None, kwargs=None, name=None):
        """Run task manager every hour.
        
        Args:
            time: str, '30:27' means the task will start at the 30th minute and 27 seconds every hour.
            function: task function.
            args: list, function args.
            kwargs: dict, function kwargs.
            name: task name, if None, is function name.
        """
        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()
        if name is None:
            name = function.__name__+(f'_{len(self.config)+1}' if function.__name__ in self.config else '')
        self.config[name] = {'mode':'every_hour', 'time':time, 'function':function, 'args':args, 
                             'kwargs':kwargs, 'execute_num':0, 'runner':(function, args, kwargs, name),
                             'time_init':datetime.datetime.now()}
        self.params.tracker_dict[name] = dict()
    
    def every_day(self, time, function, args=None, kwargs=None, name=None):
        """Run task manager every day.
        
        Args:
            time: str, '08:30:27' means that the task will start at the 30th minute and 27 seconds of the 8th hour every day.
            function: task function.
            args: list, function args.
            kwargs: dict, function kwargs.
            name: task name, if None, is function name.
        """
        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()
        if name is None:
            name = function.__name__+(f'_{len(self.config)+1}' if function.__name__ in self.config else '')
        self.config[name] = {'mode':'every_day', 'time':time, 'function':function, 'args':args, 
                             'kwargs':kwargs, 'execute_num':0, 'runner':(function, args, kwargs, name),
                             'time_init':datetime.datetime.now()}
        self.params.tracker_dict[name] = dict()
    
    def every_week(self, time, function, args=None, kwargs=None, name=None):
        """Run task manager every week.
        
        Args:
            time: str, '1,3,7:08:30:27' means that every Monday, Wednesday, Sunday, 
                  the 8th hour, 30 minutes and 27 seconds to start the task.
            function: task function.
            args: list, function args.
            kwargs: dict, function kwargs.
            name: task name, if None, is function name.
        """
        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()
        if name is None:
            name = function.__name__+(f'_{len(self.config)+1}' if function.__name__ in self.config else '')
        self.config[name] = {'mode':'every_week', 'time':time, 'function':function, 'args':args, 
                             'kwargs':kwargs, 'execute_num':0, 'runner':(function, args, kwargs, name),
                             'time_init':datetime.datetime.now()}
        self.params.tracker_dict[name] = dict()
    
    def every_month(self, time, function, args=None, kwargs=None, name=None):
        """Run task manager every month.
        
        Args:
            time: str, '1,13,27:08:30:27' means that the task will start at the 30th minute and 27 seconds 
                  of the 8th hour on the 1st, 13th, and 27th of each month.
            function: task function.
            args: list, function args.
            kwargs: dict, function kwargs.
            name: task name, if None, is function name.
        """
        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()
        if name is None:
            name = function.__name__+(f'_{len(self.config)+1}' if function.__name__ in self.config else '')
        self.config[name] = {'mode':'every_month', 'time':time, 'function':function, 'args':args, 
                             'kwargs':kwargs, 'execute_num':0, 'runner':(function, args, kwargs, name),
                             'time_init':datetime.datetime.now()}
        self.params.tracker_dict[name] = dict()
    
    def _reset_time(self, name, time_now=None):
        if time_now is None:
            time_now = datetime.datetime.now()
        if self.config[name]['mode']=='every_minute':
            seconds = 60+self.config[name]['time']-time_now.second
        elif self.config[name]['mode']=='every_hour':
            split = self.config[name]['time'].split(':')
            seconds = int(datetime.datetime(time_now.year, time_now.month, time_now.day, time_now.hour, 
                                            int(split[0]), int(split[1]), time_now.microsecond
                                           ).timestamp()-time_now.timestamp())
            if seconds<40:
                seconds = 3600+seconds
        elif self.config[name]['mode']=='every_day':
            split = self.config[name]['time'].split(':')
            seconds = int(datetime.datetime(time_now.year, time_now.month, time_now.day, int(split[0]), 
                                            int(split[1]), int(split[2]), time_now.microsecond
                                           ).timestamp()-time_now.timestamp())
            if seconds<40:
                seconds = 86400+seconds
        elif self.config[name]['mode']=='every_week':
            split = self.config[name]['time'].split(':')
            seconds = [int(i)-time_now.weekday()-1 for i in split[0].split(',')]
            seconds = [(datetime.datetime(time_now.year, time_now.month, time_now.day, int(split[1]), 
                                      int(split[2]), int(split[3]), time_now.microsecond
                                     )+datetime.timedelta(days=7+i if i<0 else i)
                    ).timestamp()-time_now.timestamp() 
                    for i in seconds]
            if max(seconds)<40:
                seconds= 604800+min(seconds)
            else:
                seconds = [i for i in sorted(seconds) if i>=40][0]
        elif self.config[name]['mode']=='every_month':
            split = self.config[name]['time'].split(':')

            seconds = [datetime.datetime(time_now.year, time_now.month, int(i), int(split[1]), 
                                      int(split[2]), int(split[3]), time_now.microsecond
                                     ) for i in split[0].split(',')]
            if time_now.month<12:
                seconds += [datetime.datetime(time_now.year, time_now.month+1, int(i), int(split[1]), 
                                      int(split[2]), int(split[3]), time_now.microsecond
                                     ) for i in split[0].split(',')]
            else:
                seconds += [datetime.datetime(time_now.year+1, 1, int(i), int(split[1]), 
                                      int(split[2]), int(split[3]), time_now.microsecond
                                     ) for i in split[0].split(',')]
            seconds = [i.timestamp()-time_now.timestamp() for i in seconds]
            seconds = [i for i in sorted(seconds) if i>=40][0]

        self.config[name]['time_next'] = time_now+datetime.timedelta(seconds=seconds)
        self.config[name]['time_record'] = time_now
    
    def run(self):
        time_now = datetime.datetime.now()
        if self.params.config_file is not None:
            config = Config(file_py=self.params.config_file)
            for name in config.config:
                self.config[name] = config.config[name]
                self.config[name]['execute_num'] = 0
                self.config[name]['runner'] = (self.config[name]['function'], self.config[name]['args'], 
                                               self.config[name]['kwargs'], name)
                self.config[name]['time_init'] = time_now
                
        for name in self.config:
            self._reset_time(name, time_now)
            if self.params.logger.params.log_file!='':
                self.params.logger.write(f'New task {name} has been added.')
            if self.params.verbose:
                self.params.logger.info(f'New task {name} has been added.')
        
        while True:
            time_now = datetime.datetime.now()
            
            for name in self.config:
                if self.config[name]['time_next']>time_now:
                    self.config[name]['time_record'] = time_now
                else:
                    self._start(self.config[name]['runner'])                    
                    self._reset_time(name, time_now)
                    self.config[name]['execute_num'] += 1
            
            try:
                if self.params.config_file is not None:
                    config = Config(file_py=self.params.config_file)
                    for name in config.config:
                        if name not in self.config:
                            self.config[name] = config.config[name]
                            self.config[name]['execute_num'] = 0
                            self.config[name]['time_init'] = time_now
                            self._reset_time(name, time_now)
                            if self.params.logger.params.log_file!='':
                                self.params.logger.write(f'New task {name} has been added.')
                            if self.params.verbose:
                                self.params.logger.info(f'New task {name} has been added.')
                        for i,j in config.config[name].items():
                            self.config[name][i] = j
                        self.config[name]['runner'] = (self.config[name]['function'], self.config[name]['args'], 
                                                       self.config[name]['kwargs'], name)
            except Exception as msg:
                if self.params.logger.params.log_file!='':
                    self.params.logger.write(str(msg))
                if self.params.verbose:
                    self.params.logger.info(str(msg))
    
    def _run(self, runner):
        """Runs function runner """
        output, error, got_error = None, None, False
        started = datetime.datetime.now()
        try:
            output = runner[0](*runner[1], **runner[2])
        except Exception as e:
            got_error = True
            error = str(e)
        finally:
            finished = datetime.datetime.now()
            self.params.tracker_dict[runner[3]] = {
                "output": output,
                "started_time": started,
                "finished_time": finished,
                "execution_time": (finished - started).total_seconds(),
                "got_error": got_error,
                "error": error}
        msg = (f'task {runner[3]} started_time: {str(started)[:19]}, finished_time: {str(finished)[:19]}, '+
               f'execution_time: {round((finished - started).total_seconds(),4)}s, got_error: {got_error}, error: {error}.')
        if self.params.logger.params.log_file!='':
            self.params.logger.write(msg)
        if self.params.verbose:
            self.params.logger.info(msg)
            
    def _start(self, runner):
        """Starts runner process """
        self.params.runner_dict[runner[3]] = multiprocessing.Process(target=self._run, args=(runner,))
        self.params.runner_dict[runner[3]].daemon = True
        self.params.runner_dict[runner[3]].start()



from linora.utils._config import Config

__all__ = ['EarlyStopping']


class EarlyStopping():
    """Stop training when a monitored metric has stopped improving.
    
    Args:
        monitor: Quantity to be monitored.
        min_delta: Minimum change in the monitored quantity to qualify as an improvement, 
            i.e. an absolute change of less than min_delta, will count as no improvement.
        patience: Number of batch with no improvement after which training will be stopped.
        mode: One of {"min", "max"}. In min mode, training will stop when the quantity monitored has stopped decreasing; 
            in "max" mode it will stop when the quantity monitored has stopped increasing.
        baseline: Baseline value for the monitored quantity. 
            Training will stop if the model doesn't show improvement over the baseline.
    """
    def __init__(self, monitor, min_delta=0, patience=0, mode='min', baseline=None):
        self._params = Config()
        self._params.monitor = monitor
        self._params.min_delta = min_delta
        self._params.patience = patience
        self._params.mode = mode
        self._params.baseline = baseline
        self._params.history = []
        self._params.state = False
        self._params.name = 'EarlyStopping'
        
    def _update(self, batch, log):
        """update log.
        
        Args:
            batch: Integer, index of batch.
            log: dict, name and value of loss or metrics;
        """
        if self._params.monitor in log:
            self._params.history = self._params.history[-self._params.patience:]+[log[self._params.monitor]]
            if self._params.mode=='min':
                self._params.state = min(self._params.history[-self._params.patience:])+self._params.min_delta>self._params.history[0]
                if not self._params.state and self._params.baseline is not None:
                    self._params.state = min(self._params.history[-self._params.patience:])>self._params.baseline
            else:
                self._params.state = max(self._params.history[-self._params.patience:])-self._params.min_delta<self._params.history[0]
                if not self._params.state and self._params.baseline is not None:
                    self._params.state = max(self._params.history[-self._params.patience:])<self._params.baseline
        


from linora.feature_column._datetime import *


class FeatureDatetime(object):
    def __init__(self):
        self.pipe = {}
        
    def datetime_extraction(self, variable, extract, name=None, keep=True):
        """Extracts date and time features from datetime variables.

        The transformer supports the extraction of the following features:
        - 'year'
        - 'month'
        - 'day'
        - 'hour'
        - 'minute'
        - 'second'
        - 'date'
        - 'time'
        - 'weekday'
        - 'yearday'
        - 'quarter'
        - 'week'
        - 'days_in_month'
        - 'month_start'
        - 'month_end'
        - 'quarter_start'
        - 'quarter_end'
        - 'year_start'
        - 'year_end'
        - 'leap_year'

        Args:
            variable: str, feature variable name.
            extract: The list of date features to extract. 
            name: str, output feature name, if None, name is feature.name .
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'extract':extract, 'name':variable if name is None else name},
                  'type':'datetime_extraction', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self
    
    def datetime_transform(self, variable, format_in, format_out='timestamp', name=None, keep=True):
        """Time format transform.

        The transformer supports the time format of the following features:

        - 'timenumber:s', 's' id unit, one of ['s', 'ms', 'us', 'ns']:
            1679455111, 1679455111.0958722
        - 'strtime:format', format is "%Y-%m-%d %H:%M:%S":
            %y Two-digit Year Representation (00-99)
            %Y Four-digit Year Representation (000-9999)
            %m month (01-12)
            %d One day in months (0-31)
            %h Hours in 24-hour clock (0-23)
            %I 12-hour Hours (01-12)
            %M minutes (00=59)
            %S seconds (00-59)
            %a local simplified week name
            %A local full week name
            %b Local Simplified Month Name
            %B Local full month name
            %c local corresponding date representation and time representation
            %j One day in year (001-366)
            %p equivalent of local a.m. or p.m.
            %U Number of weeks in a year (00-53) Sunday is the beginning of the week.
            %w week (0-6), Sunday is the beginning of the week.
            %W Number of weeks in a year (00-53) Monday is the beginning of the week.
            %x local corresponding date representation
            %X local corresponding time representation
            %Z Name of the current time zone
            eg:
            - "%Y-%m-%d %H:%M:%S": '2023-01-01 02:03:56'
            - "%A %Y-%m %H:%M:%S": 'Thursday 2023-01 02:03:56'
        - 'timestamp':
            Timestamp(2023-01-01 02:03:56)
        - 'datetime'
            datetime.datetime(2023, 3, 23, 8, 9, 59, 733760)

        Args:
            variable: str, feature variable name.
            format_in: input time format.
            format_out: output time format.
            name: str, output feature name, if None, name is feature.name .
            keep: If the `name` is output only once in the calculation, the `name` will be kept in the final result.
        """
        config = {'param':{'format_in':format_in, 'format_out':format_out,
                           'name':variable if name is None else name},
                  'type':'datetime_transform', 'variable':variable, 'keep':keep}
        self.pipe[len(self.pipe)] = config
        return self

